import pytest
from flask import Flask
from mock import patch

from gordo.server.prometheus import GordoServerPrometheusMetrics
from prometheus_client import CollectorRegistry


@pytest.fixture
def registry():
    return CollectorRegistry()


@pytest.fixture
def prometheus_metrics(registry):
    return GordoServerPrometheusMetrics(
        args_labels=(("gordo_project", "project"), ("gordo_name", "model")),
        info={"version": "0.60.0"},
        ignore_paths=["/healthcheck"],
        registry=registry,
    )


@pytest.fixture
def gordo_server(prometheus_metrics: GordoServerPrometheusMetrics):
    app = Flask("test_server")

    @app.route("/healthcheck")
    def healthcheck():
        return "", 200

    @app.route("/success/<gordo_project>/<gordo_name>")
    def success(gordo_project: str, gordo_name: str):
        return "", 200

    @app.route("/failed/<gordo_project>")
    def failed(gordo_project: str):
        return "", 500

    with patch("gordo.server.prometheus.current_time", side_effect=[0.0, 0.25]):
        prometheus_metrics.prepare_app(app)

        yield app


def test_success(gordo_server: Flask, registry: CollectorRegistry):
    client = gordo_server.test_client()
    client.get("/success/project1/gordo1")
    sample_value = registry.get_sample_value("gordo_server_info", {"version": "0.60.0"})
    assert sample_value == 1.0, "Metric gordo_server_info != 1.0"
    sample_value = registry.get_sample_value(
        "gordo_server_request_duration_seconds_bucket",
        {
            "version": "0.60.0",
            "project": "project1",
            "model": "gordo1",
            "method": "GET",
            "path": "/success/<gordo_project>/<gordo_name>",
            "status_code": "200",
            "le": "0.25",
        },
    )
    assert (
        sample_value == 1.0
    ), "Metric gordo_server_request_duration_seconds_bucket != 1.0"
    sample_value = registry.get_sample_value(
        "gordo_server_requests_total",
        {
            "version": "0.60.0",
            "project": "project1",
            "model": "gordo1",
            "method": "GET",
            "path": "/success/<gordo_project>/<gordo_name>",
            "status_code": "200",
        },
    )
    assert sample_value == 1.0, "Metric gordo_server_requests_total != 1.0"


def test_failed(gordo_server: Flask, registry: CollectorRegistry):
    client = gordo_server.test_client()
    client.get("/failed/project1")
    sample_value = registry.get_sample_value(
        "gordo_server_request_duration_seconds_bucket",
        {
            "version": "0.60.0",
            "project": "project1",
            "model": "",
            "method": "GET",
            "path": "/failed/<gordo_project>",
            "status_code": "500",
            "le": "0.25",
        },
    )
    assert (
        sample_value == 1.0
    ), "Metric gordo_server_request_duration_seconds_bucket != 1.0"
    sample_value = registry.get_sample_value(
        "gordo_server_requests_total",
        {
            "version": "0.60.0",
            "project": "project1",
            "model": "",
            "method": "GET",
            "path": "/failed/<gordo_project>",
            "status_code": "500",
        },
    )
    assert sample_value == 1.0, "Metric gordo_server_requests_total != 1.0"


def test_ignore(gordo_server: Flask, registry: CollectorRegistry):
    client = gordo_server.test_client()
    client.get("/healthcheck")
    sample_value = registry.get_sample_value(
        "gordo_server_requests_total",
        {
            "version": "0.60.0",
            "project": "",
            "model": "",
            "method": "GET",
            "path": "/healthcheck",
            "status_code": "200",
        },
    )
    assert sample_value is None, "Metric gordo_server_requests_total is not None"



import sys
import json
import pytest

from io import StringIO

from gordo.cli.exceptions_reporter import (
    ReportLevel,
    ExceptionsReporter,
    DEFAULT_EXIT_CODE,
)


def test_report_level():
    level = ReportLevel.get_by_name("MESSAGE")
    assert level == ReportLevel.MESSAGE
    level = ReportLevel.get_by_name("DIFFERENT")
    assert level is None
    level = ReportLevel.get_by_name("DIFFERENT", ReportLevel.MESSAGE)
    assert level == ReportLevel.MESSAGE
    levels = ReportLevel.get_names()
    assert len(levels) == 4


class _Test1Exception(Exception):
    pass


class _Test2Exception(Exception):
    pass


class _Test3Exception(_Test1Exception):
    pass


traceback_example = [
    "Traceback (most recent call last):\n",
    '  File "bar.py", line 13, in <module>\n    foo()\n',
    '  File "bar.py", line 10, in foo\n    bar()\n',
    '  File "bar.py", line 7, in bar\n'
    '    raise Exception("Something bad happening")\n',
    "Exception: Something bad happening\n",
]


@pytest.fixture
def reporter1():
    return ExceptionsReporter(((_Test1Exception, 110),))


def test_sort_exceptions():
    exceptions = ((Exception, 10), (IOError, 20), (FileNotFoundError, 30))
    sorted_exceptions = ExceptionsReporter.sort_exceptions(exceptions)
    assert sorted_exceptions == [
        (FileNotFoundError, 30),
        (OSError, 20),
        (Exception, 10),
    ]


def test_trim_formatted_traceback():
    result = ExceptionsReporter.trim_formatted_traceback(traceback_example, 118)
    assert result == [
        "...\n",
        '  File "bar.py", line 7, in bar\n'
        '    raise Exception("Something bad happening")\n',
        "Exception: Something bad happening\n",
    ]
    result = ExceptionsReporter.trim_formatted_traceback(traceback_example, 117)
    assert result == ["...\n", "Exception: Something bad happening\n"]


def test_reporter1(reporter1):
    assert reporter1.exception_exit_code(_Test1Exception) == 110
    assert reporter1.exception_exit_code(_Test2Exception) == DEFAULT_EXIT_CODE
    assert reporter1.exception_exit_code(_Test3Exception) == 110
    assert reporter1.exception_exit_code(None) == 0


def test_reporting_out_of_exception_scope(reporter1):
    sio = StringIO()
    reporter1.report(ReportLevel.MESSAGE, None, None, None, sio)
    assert json.loads(sio.getvalue()) == {}


def report(e, reporter, report_level, report_file, **kwargs):
    try:
        raise e
    except Exception:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        reporter.report(
            report_level, exc_type, exc_value, exc_traceback, report_file, **kwargs
        )


def report_to_string(e, reporter, report_level, **kwargs):
    sio = StringIO()
    report(e, reporter, report_level, sio, **kwargs)
    value = sio.getvalue()
    return json.loads(value)


def test_with_message_report_level(reporter1):
    result = report_to_string(
        _Test1Exception("Test message"), reporter1, ReportLevel.MESSAGE
    )
    assert result == {"type": "_Test1Exception", "message": "Test message"}


def test_with_traceback_report_level(reporter1):
    result = report_to_string(
        _Test1Exception("Test message"), reporter1, ReportLevel.TRACEBACK
    )
    assert result["type"] == "_Test1Exception"
    assert "traceback" in result
    assert "Test message" in result["traceback"]


def test_with_type_report_level(reporter1):
    result = report_to_string(
        _Test1Exception("Test message"), reporter1, ReportLevel.TYPE
    )
    assert result == {"type": "_Test1Exception"}


def test_with_exit_code_report_level(reporter1):
    result = report_to_string(
        _Test1Exception("Test message"), reporter1, ReportLevel.EXIT_CODE
    )
    assert result == {}


def test_with_unicode_chars(reporter1):
    result = report_to_string(
        _Test1Exception("\t你好 world!\n"), reporter1, ReportLevel.MESSAGE
    )
    assert result == {"type": "_Test1Exception", "message": "\t?? world!\n"}


def test_with_max_message_len(reporter1):
    result = report_to_string(
        _Test1Exception("Hello world!"),
        reporter1,
        ReportLevel.MESSAGE,
        max_message_len=8,
    )
    assert result == {"type": "_Test1Exception", "message": "Hello..."}
    result = report_to_string(
        _Test1Exception("Hello world!"),
        reporter1,
        ReportLevel.MESSAGE,
        max_message_len=20,
    )
    assert result == {"type": "_Test1Exception", "message": "Hello world!"}
    result = report_to_string(
        _Test1Exception("Hello"), reporter1, ReportLevel.MESSAGE, max_message_len=4
    )
    assert result == {"type": "_Test1Exception", "message": ""}



# -*- coding: utf-8 -*-

import unittest
import logging
import json
import os

from tempfile import TemporaryDirectory

import pytest
import numpy as np
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.decomposition import TruncatedSVD, PCA
from sklearn.preprocessing import MinMaxScaler
from sklearn.compose import TransformedTargetRegressor
from gordo.machine.model.anomaly.diff import DiffBasedAnomalyDetector

from gordo.machine.model.models import KerasAutoEncoder
from gordo import serializer


logger = logging.getLogger(__name__)


class PipelineSerializationTestCase(unittest.TestCase):
    def test_pipeline_serialization(self):

        pipe = Pipeline(
            [
                ("pca1", PCA(n_components=10)),
                (
                    "fu",
                    FeatureUnion(
                        [
                            ("pca2", PCA(n_components=3)),
                            (
                                "pipe",
                                Pipeline(
                                    [
                                        ("minmax", MinMaxScaler()),
                                        ("truncsvd", TruncatedSVD(n_components=7)),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
            ]
        )

        X = np.random.random(size=100).reshape(10, 10)
        pipe.fit(X.copy(), X.copy())

        with TemporaryDirectory() as tmp:

            # Test dump
            metadata = {"key": "value"}
            serializer.dump(pipe, tmp, metadata=metadata)

            # Test load from the serialized pipeline above
            pipe_clone = serializer.load(tmp)
            metadata_clone = serializer.load_metadata(tmp)

            # Ensure the metadata was saved and loaded back
            self.assertEqual(metadata, metadata_clone)

            # Verify same state for both pipelines
            y_hat_pipe1 = pipe.predict(X.copy()).flatten()
            y_hat_pipe2 = pipe_clone.predict(X.copy()).flatten()
            self.assertTrue(np.allclose(y_hat_pipe1, y_hat_pipe2))

            # Now use dumps/loads
            serialized = serializer.dumps(pipe)
            pipe_clone = serializer.loads(serialized)

            # Verify same state for both pipelines
            y_hat_pipe1 = pipe.predict(X.copy()).flatten()
            y_hat_pipe2 = pipe_clone.predict(X.copy()).flatten()
            self.assertTrue(np.allclose(y_hat_pipe1, y_hat_pipe2))


@pytest.mark.parametrize(
    "model",
    [
        KerasAutoEncoder(kind="feedforward_hourglass"),
        DiffBasedAnomalyDetector(
            base_estimator=TransformedTargetRegressor(
                regressor=KerasAutoEncoder(kind="feedforward_symmetric"),
                transformer=MinMaxScaler(),
            )
        ),
        TransformedTargetRegressor(
            regressor=Pipeline(
                steps=[
                    ("stp1", MinMaxScaler()),
                    ("stp2", KerasAutoEncoder(kind="feedforward_symmetric")),
                ]
            )
        ),
    ],
)
def test_dump_load_models(model):

    X = np.random.random(size=100).reshape(10, 10)
    model.fit(X.copy(), X.copy())
    model_out = model.predict(X.copy())

    with TemporaryDirectory() as tmp:
        serializer.dump(model, tmp)

        model_clone = serializer.load(tmp)
        model_clone_out = model_clone.predict(X.copy())

        assert np.allclose(model_out.flatten(), model_clone_out.flatten())


@pytest.mark.parametrize("location", ("metadata.json", "../metadata.json", None))
def test_load_metadata(tmpdir, location):
    """
    Test load_metadata can look in directory given as well as directory above that
    along with dealing with 'FileNotFoundError' when a non-existent file is given.
    """
    model_dir = os.path.join(tmpdir, "some-model-dir")
    os.mkdir(model_dir)
    if location:
        with open(os.path.join(model_dir, location), "w") as f:
            json.dump(dict(key="value"), f)
        assert serializer.load_metadata(model_dir) == dict(key="value")
    else:
        # Attempting to load a file which doesn't exist will raise FileNotFoundError
        with pytest.raises(FileNotFoundError):
            assert serializer.load_metadata(tmpdir)



# -*- coding: utf-8 -*-

import datetime

import pytest

from gordo.machine.validators import (
    ValidUrlString,
    ValidDatetime,
    ValidTagList,
    ValidMetadata,
    ValidModel,
    ValidMachineRuntime,
    fix_resource_limits,
)


def test_valid_model():
    """
    Verifies a given object is Union[str, dict]
    """

    class MyClass:
        value = ValidModel()

    myclass = MyClass()

    # All ok
    myclass.value = {
        "gordo.machine.model.models.KerasAutoEncoder": {"kind": "feedforward_hourglass"}
    }
    myclass.value = "sklearn.ensemble.RandomForestRegressor"

    # Not ok
    with pytest.raises(ValueError):
        myclass.value = 1
    with pytest.raises(ValueError):
        myclass.value = None


def test_valid_metadata():
    """
    Verifies a given object is Optional[dict]
    """

    class MyClass:
        value = ValidMetadata()

    myclass = MyClass()

    # All of these are ok
    myclass.value = dict()
    myclass.value = {"key": "value"}
    myclass.value = None

    # This is not
    with pytest.raises(ValueError):
        myclass.value = 1
    with pytest.raises(ValueError):
        myclass.value = "string"


def test_valid_datetime():
    """
    Verifies a given object is of datetime.datetime
    """

    class MyClass:
        value = ValidDatetime()

    myclass = MyClass()

    myclass.value = datetime.datetime.now(tz=datetime.timezone.utc)
    # Should raise an error
    with pytest.raises(ValueError):
        myclass.value = datetime.datetime.now()

    # Should raise an error
    with pytest.raises(ValueError):
        myclass.value = "not a datetime object"


def test_valid_tag_list():
    """
    Verifies a given object is of a non-empty List[str]
    """

    class MyClass:
        value = ValidTagList()

    myclass = MyClass()

    # No problem
    myclass.value = ["string here", "string there"]

    # Problem.
    with pytest.raises(ValueError):
        myclass.value = "not a list"
    with pytest.raises(ValueError):
        myclass.value = []  # Error with empty list.


def test_valid_url_string():
    """
    Verifies that a given string matches what we consider to be a valid
    url; in this case alphanumeric with dashes.
    """
    valid_names = [
        "valid-name-here",
        "validnamehere",
        "also-a-valid-name123",
        "equally-valid-name",
        "another-1-2-3",
    ]
    for valid_name in valid_names:
        assert ValidUrlString.valid_url_string(
            valid_name
        ), f"Expected '{valid_name}' to be valid, but it is not"

    invalid_names = [
        "Not_a_valid_name",
        "C%tainly-not-v@lid",
        "also not valid with spaces",
        "couldn't-possibly-valid",
        "also,-this-is-not-valid",
        "(this-is-not-either)",
        "!nor-this",
        "cannot-have-UpperCase",
        "-cannot-begin-with-dashes",
        "cannot-end-with-dashes-",
    ]
    for invalid_name in invalid_names:
        assert not ValidUrlString.valid_url_string(
            invalid_name
        ), f"Expected '{invalid_name}' to be invalid, but it is."

    # Test it's a working Descriptor
    class MyClass:
        value = ValidUrlString()

    myclass = MyClass()

    myclass.value = "this-should-not-cause-an-error"

    with pytest.raises(ValueError):
        myclass.value = (
            "this-name-is-simply-too-long-that-is-the-only-problem-here-for-sure"
        )

    with pytest.raises(ValueError):
        myclass.value = "but this should!"


def test_valid_runtime():
    """
    Verifies that ValidRuntime enforces a server section with resources.requests and
    resources.limits
    """

    class MyClass:
        value = ValidMachineRuntime()

    myclass = MyClass()

    # Runtimes should be dicts
    myclass.value = {"influx": {"enable": True}}

    # Problem.
    with pytest.raises(ValueError):
        myclass.value = 2


def test_fix_resource_limits():
    # Bumps cpu limit
    resource_dict = {
        "requests": {"memory": 1, "cpu": 5},
        "limits": {"memory": 3, "cpu": 4},
    }
    res = fix_resource_limits(resource_dict)
    assert res["limits"]["cpu"] == 5

    # Works without requests
    resource_dict = {"limits": {"memory": 3, "cpu": 4}}
    res = fix_resource_limits(resource_dict)
    assert res["limits"]["cpu"] == 4
    # Works without limits
    resource_dict = {"requests": {"memory": 1, "cpu": 5}}
    res = fix_resource_limits(resource_dict)
    assert res["requests"]["cpu"] == 5

    # Does  nothing if limit>requests
    resource_dict = {
        "requests": {"memory": 1, "cpu": 5},
        "limits": {"memory": 3, "cpu": 6},
    }
    res = fix_resource_limits(resource_dict)
    assert res["limits"]["cpu"] == 6


def test_fix_resource_limits_checks_ints():
    resource_dict = {
        "requests": {"memory": "1M", "cpu": 5},
        "limits": {"memory": 3, "cpu": 4},
    }
    with pytest.raises(ValueError):
        fix_resource_limits(resource_dict)



# -*- coding: utf-8 -*-

import pathlib

from gordo.util import disk_registry


def test_simple_happy_path(tmpdir):
    """Tests that it works to write and read a simple value to a fresh registry"""
    disk_registry.write_key(tmpdir, "akey", "aval")
    assert disk_registry.get_value(tmpdir, "akey") == "aval"


def test_new_registry(tmpdir):
    """
    Tests that it works to write and read a simple value to a fresh registry with
    a non-existing directory
    """
    registry = pathlib.Path(tmpdir).joinpath("newregistry")
    disk_registry.write_key(registry, "akey", "aval")
    assert disk_registry.get_value(registry, "akey") == "aval"


def test_complicated_happy_path(tmpdir):
    """Tests that it works to write and read a 'complicated' value"""
    value = """
    A long
    value with many weird character lie åøæ
    and some linebreaks"""
    disk_registry.write_key(tmpdir, "akey", value)
    assert disk_registry.get_value(tmpdir, "akey") == value


def test_overwrites_existing(tmpdir):
    """Double writes to the same registry overwrites the first value"""
    the_key = "akey"
    first_value = "Some value"
    disk_registry.write_key(tmpdir, the_key, first_value)
    assert disk_registry.get_value(tmpdir, the_key) == first_value
    second_value = "Some value"
    disk_registry.write_key(tmpdir, the_key, second_value)
    assert disk_registry.get_value(tmpdir, the_key) == second_value


def test_delete(tmpdir):
    """Delete removes a key"""
    the_key = "akey"
    first_value = "Some value"
    disk_registry.write_key(tmpdir, the_key, first_value)
    assert disk_registry.get_value(tmpdir, the_key) == first_value

    existed_p = disk_registry.delete_value(tmpdir, the_key)
    assert disk_registry.get_value(tmpdir, the_key) is None
    # They key existed
    assert existed_p


def test_double_delete(tmpdir):
    """Delete works on non-existing key, returning False"""
    the_key = "akey"
    existed_p = disk_registry.delete_value(tmpdir, the_key)
    assert disk_registry.get_value(tmpdir, the_key) is None
    assert not existed_p


def test_get_value_without_registry_dir():
    """Test that it works to have registry_dir as None, and that it returns None as expected"""
    assert disk_registry.get_value(None, "akey") is None



import json
import traceback

from typing import Tuple, Iterable, Type, IO, Optional, List, Dict
from types import TracebackType
from collections import Counter
from enum import Enum

from gordo.util import replace_all_non_ascii_chars


class ReportLevel(Enum):
    EXIT_CODE = 0
    TYPE = 1
    MESSAGE = 2
    TRACEBACK = 3

    @classmethod
    def get_by_name(
        cls, name: str, default: Optional["ReportLevel"] = None
    ) -> Optional["ReportLevel"]:
        for level in cls:
            if name == level.name:
                return level
        return default

    @classmethod
    def get_names(cls):
        return [level.name for level in cls]


DEFAULT_EXIT_CODE = 1


class ExceptionsReporter:
    """
        Helper which can save the exception information in JSON format into the file.
        This class might be used for storing exception information after termination of the Kubernetes pod.
        `Information <https://kubernetes.io/docs/tasks/debug-application-cluster/determine-reason-pod-failure/#customizing-the-termination-message>`_
    """

    def __init__(
        self,
        exceptions: Iterable[Tuple[Type[Exception], int]],
        default_exit_code: int = DEFAULT_EXIT_CODE,
        traceback_limit: Optional[int] = None,
    ):
        """
        Parameters
        ----------
        exceptions
            Exceptions list with preferred exit codes for each of them
        default_exit_code
            Default exit code. It might be used as `sys.exit()` code
        traceback_limit
            Limit for `traceback.format_exception()`
        """
        self.exceptions_items = self.sort_exceptions(exceptions)
        self.default_exit_code = default_exit_code
        self.traceback_limit = traceback_limit

    @staticmethod
    def sort_exceptions(
        exceptions: Iterable[Tuple[Type[Exception], int]]
    ) -> List[Tuple[Type[Exception], int]]:
        inheritance_levels: Dict[Type[BaseException], int] = Counter()
        for exc, exit_code in exceptions:
            for e, _ in exceptions:
                if e is not exc and issubclass(exc, e):
                    inheritance_levels[e] += 1
        sorted_exceptions = list(exceptions)

        def key(v):
            exc, exit_code = v
            return (inheritance_levels[exc], exit_code)

        return sorted(sorted_exceptions, key=key)

    @staticmethod
    def trim_message(message: str, max_length: int) -> str:
        if len(message) > max_length:
            message = message[: max_length - 3]
            return "" if len(message) <= 3 else message + "..."
        return message

    @staticmethod
    def trim_formatted_traceback(
        formatted_traceback: List[str], max_length: int
    ) -> List[str]:
        if sum(len(line) for line in formatted_traceback) <= max_length:
            return formatted_traceback
        length = 4
        result = []
        for line in reversed(formatted_traceback):
            length += len(line)
            if length > max_length:
                result.append("...\n")
                break
            else:
                result.append(line)
        return list(reversed(result))

    def found_exception_item(self, exc_type: Type[BaseException]):
        for item in self.exceptions_items:
            if issubclass(exc_type, item[0]):
                return item
        return None

    def exception_exit_code(self, exc_type: Optional[Type[BaseException]]) -> int:
        """
        Possible `sys.exit()` code for given exception type

        Parameters
        ----------
        exc_type
            The exception type

        Returns
        -------
        int
        """
        if exc_type is None:
            return 0
        item = self.found_exception_item(exc_type)
        return item[1] if item is not None else self.default_exit_code

    def report(
        self,
        level: ReportLevel,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
        report_file: IO[str],
        max_message_len: Optional[int] = None,
    ):
        """
        Report exception to the file.
        `exc_type`, `exc_value`, `exc_traceback` might be values returned by `sys.exc_info()`

        Parameters
        ----------
        level: ReportLevel
            Level of the report verbosity
        exc_type
            The exception type
        exc_value
            The exception
        exc_traceback
            The exception traceback
        report_file
            File like object for reporting
        max_message_len
            The maximum length of `message` or `traceback`.
            Actual for the environments with limitation of storage capacity.
            For example, 2024 bytes is the maximum size of the k8s pod termination message content
        """
        report = {}

        if exc_type is not None and exc_value is not None and exc_traceback is not None:
            if self.found_exception_item(exc_type) is not None:
                if level in (
                    ReportLevel.MESSAGE,
                    ReportLevel.TYPE,
                    ReportLevel.TRACEBACK,
                ):
                    report["type"] = replace_all_non_ascii_chars(exc_type.__name__, "?")
                if level == ReportLevel.MESSAGE:
                    report["message"] = replace_all_non_ascii_chars(str(exc_value), "?")
                    if max_message_len is not None:
                        report["message"] = self.trim_message(
                            report["message"], max_message_len
                        )
                elif level == ReportLevel.TRACEBACK:
                    formatted_traceback = traceback.format_exception(
                        exc_type, exc_value, exc_traceback, limit=self.traceback_limit
                    )
                    formatted_traceback = [
                        replace_all_non_ascii_chars(v, "?") for v in formatted_traceback
                    ]
                    if max_message_len is not None:
                        formatted_traceback = self.trim_formatted_traceback(
                            formatted_traceback, max_message_len
                        )
                    report["traceback"] = "".join(formatted_traceback)
        json.dump(report, report_file)

    def safe_report(
        self,
        level: ReportLevel,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
        report_file_path: str,
        max_message_len: Optional[int] = None,
    ):
        """
        Basically this is a wrapper for `ExceptionsReporter.report()` function
        with additional internal exceptions handling

        Parameters
        ----------
        level
        exc_type
        exc_value
        exc_traceback
        report_file_path
        max_message_len

        Returns
        -------

        """
        try:
            with open(report_file_path, "w") as report_file:
                self.report(
                    level,
                    exc_type,
                    exc_value,
                    exc_traceback,
                    report_file,
                    max_message_len,
                )
        except Exception:
            traceback.print_exc()



# -*- coding: utf-8 -*-

from typing import List

import numpy as np

from tests.conftest import gordo_ml_server_client, trained_model_directory, sensors


"""
`benchmark` is a pytest-benchmark fixture: https://pytest-benchmark.readthedocs.io/en/latest/
"""


def single_post_to_ml_server(client, path: str, X: List[List[float]]):
    """Post some data under 'X' given a client and path"""
    resp = client.post(path, json={"X": X})
    return resp


def test_bench_ml_server_anomaly_post(benchmark, gordo_ml_server_client, sensors):
    """Benchmark posting data to the anomaly endpoint"""
    X = np.random.random((100, len(sensors))).tolist()
    resp = benchmark.pedantic(
        single_post_to_ml_server,
        args=(gordo_ml_server_client, "/anomaly/prediction", X),
        iterations=1,
        rounds=100,
    )
    assert resp.status_code == 200


def test_bench_ml_server_base_post(benchmark, gordo_ml_server_client, sensors):
    """Benchmark posting data to the base prediction endpoint"""
    X = np.random.random((100, len(sensors))).tolist()
    resp = benchmark.pedantic(
        single_post_to_ml_server,
        args=(gordo_ml_server_client, "/prediction", X),
        iterations=1,
        rounds=100,
    )
    assert resp.status_code == 200



import logging
import time
import pkg_resources
import os

from typing import Dict, Any, TypeVar, Type, List, Tuple, cast

import click
import json
from jinja2 import Environment, BaseLoader

from pydantic import parse_obj_as, ValidationError
from gordo import __version__
from gordo.workflow.config_elements.normalized_config import NormalizedConfig
from gordo.workflow.workflow_generator import workflow_generator as wg
from gordo.cli.exceptions_reporter import ReportLevel
from gordo.util.version import parse_version


logger = logging.getLogger(__name__)


PREFIX = "WORKFLOW_GENERATOR"
DEFAULT_BUILDER_EXCEPTIONS_REPORT_LEVEL = ReportLevel.TRACEBACK

ML_SERVER_HPA_TYPES = ["none", "k8s_cpu", "keda"]
DEFAULT_ML_SERVER_HPA_TYPE = "k8s_cpu"

DEFAULT_KEDA_PROMETHEUS_METRIC_NAME = "gordo_server_request_duration_seconds_count"
DEFAULT_KEDA_PROMETHEUS_QUERY = 'sum(rate(gordo_server_request_duration_seconds_count{project=~"{{project_name}}",path=~".*prediction"}[30s]))'
DEFAULT_KEDA_PROMETHEUS_THRESHOLD = "1.0"


def get_builder_exceptions_report_level(config: NormalizedConfig) -> ReportLevel:
    orig_report_level = None
    try:
        orig_report_level = config.globals["runtime"]["builder"][
            "exceptions_report_level"
        ]
    except KeyError:
        pass
    if orig_report_level is not None:
        report_level = ReportLevel.get_by_name(orig_report_level)
        if report_level is None:
            raise ValueError(
                "Invalid 'runtime.builder.exceptions_report_level' value '%s'"
                % orig_report_level
            )
    else:
        report_level = DEFAULT_BUILDER_EXCEPTIONS_REPORT_LEVEL
    return report_level


T = TypeVar("T")


def parse_json(value: str, schema: Type[T]):
    try:
        data = json.loads(value)
    except json.JSONDecodeError as e:
        raise click.ClickException('Malformed JSON string: "%s"' % str(e))
    try:
        obj = parse_obj_as(schema, data)
    except ValidationError as e:
        raise click.ClickException('Schema validation error: "%s"' % str(e))
    return obj


DEFAULT_CUSTOM_MODEL_BUILDER_ENVS = """
[
    {
        "name": "DL_SERVICE_AUTH_STR",
        "valueFrom": {
            "secretKeyRef": {
                "key": "tenant_id_secret",
                "name": "dlserviceauth"
            }
        }
    },
    {
        "name": "DL2_SERVICE_AUTH_STR",
        "valueFrom": {
            "secretKeyRef": {
                "key": "tenant_id_secret",
                "name": "dl2serviceauth"
            }
        }
    }
]
"""


def validate_generate_context(context):
    if context["ml_server_hpa_type"] == "keda":
        if not context["with_keda"]:
            raise click.ClickException(
                '"--ml-server-hpa-type=keda" is only be supported with "--with-keda" flag'
            )
        if not context["prometheus_server_address"]:
            raise click.ClickException(
                '--prometheus-server-address should be specified for "--ml-server-hpa-type=keda"'
            )


KEDA_PROMETHEUS_QUERY_ARGS = ["project_name"]


def prepare_keda_prometheus_query(context):
    keda_prometheus_query = context["keda_prometheus_query"]
    if keda_prometheus_query:
        template = Environment(loader=BaseLoader).from_string(keda_prometheus_query)
        kwargs = {k: context[k] for k in KEDA_PROMETHEUS_QUERY_ARGS}
        return template.render(**kwargs)


def prepare_resources_labels(value: str) -> List[Tuple[str, Any]]:
    resources_labels: List[Tuple[str, Any]] = []
    if value:
        try:
            json_value = json.loads(value)
        except json.JSONDecodeError as e:
            raise click.ClickException(
                '"--resources-labels=%s" contains invalid JSON value: %s'
                % (value, str(e))
            )
        if isinstance(json_value, dict):
            resources_labels = cast(List[Tuple[str, Any]], list(json_value.items()))
        else:
            type_name = type(json_value).__name__
            raise click.ClickException(
                '"--resources-labels=%s" contains value with type "%s" instead "dict"'
                % (value, type_name)
            )
    return resources_labels


@click.group("workflow")
@click.pass_context
def workflow_cli(gordo_ctx):
    pass


@click.command("generate")
@click.option(
    "--machine-config",
    type=str,
    help="Machine configuration file",
    envvar=f"{PREFIX}_MACHINE_CONFIG",
    required=True,
)
@click.option("--workflow-template", type=str, help="Template to expand")
@click.option(
    "--owner-references",
    type=wg._valid_owner_ref,
    default=None,
    allow_from_autoenv=True,
    help="Kubernetes owner references to inject into all created resources. "
    "Should be a nonempty yaml/json list of owner-references, each owner-reference "
    "a dict containing at least the keys 'uid', 'name', 'kind', and 'apiVersion'",
    envvar=f"{PREFIX}_OWNER_REFERENCES",
)
@click.option(
    "--gordo-version",
    type=str,
    default=wg._docker_friendly_version(__version__),
    help="Version of gordo to use, if different than this one",
    envvar=f"{PREFIX}_GORDO_VERSION",
)
@click.option(
    "--project-name",
    type=str,
    help="Name of the project which own the workflow.",
    allow_from_autoenv=True,
    envvar=f"{PREFIX}_PROJECT_NAME",
    required=True,
)
@click.option(
    "--project-revision",
    type=str,
    default=int(time.time() * 1000),  # unix time milliseconds
    help="Revision of the project which own the workflow.",
    envvar=f"{PREFIX}_PROJECT_REVISION",
)
@click.option(
    "--output-file",
    type=str,
    required=False,
    help="Optional file to render to",
    envvar=f"{PREFIX}_OUTPUT_FILE",
)
@click.option(
    "--namespace",
    type=str,
    default="kubeflow",
    help="Which namespace to deploy services into",
    envvar=f"{PREFIX}_NAMESPACE",
)
@click.option(
    "--split-workflows",
    type=int,
    default=30,
    help="Split workflows containg more than this number of models into several "
    "workflows, where each workflow contains at most this nr of models. The "
    "workflows are outputted sequentially with '---' in between, which allows "
    "kubectl to apply them all at once.",
    envvar=f"{PREFIX}_SPLIT_WORKFLOWS",
)
@click.option(
    "--n-servers",
    type=int,
    default=None,
    help="Max number of ML Servers to use, defaults to N machines * 10",
    envvar=f"{PREFIX}_N_SERVERS",
)
@click.option(
    "--docker-repository",
    type=str,
    default="gordo",
    help="The docker repo to use for pulling component images from",
    envvar=f"{PREFIX}_DOCKER_REPOSITORY",
)
@click.option(
    "--docker-registry",
    type=str,
    default="auroradevacr.azurecr.io",  # TODO: Change to docker.io after migrating
    help="The docker registry to use for pulling component images from",
    envvar=f"{PREFIX}_DOCKER_REGISTRY",
)
@click.option(
    "--retry-backoff-duration",
    type=str,
    default="15s",
    help="retryStrategy.backoff.duration for workflow steps",
    envvar=f"{PREFIX}_RETRY_BACKOFF_DURATION",
)
@click.option(
    "--retry-backoff-factor",
    type=int,
    default=2,
    help="retryStrategy.backoff.factor for workflow steps",
    envvar=f"{PREFIX}_RETRY_BACKOFF_FACTOR",
)
@click.option(
    "--gordo-server-workers",
    type=int,
    help="The number of worker processes for handling Gordo server requests.",
    envvar=f"{PREFIX}_GORDO_SERVER_WORKERS",
)
@click.option(
    "--gordo-server-threads",
    type=int,
    help="The number of worker threads for handling requests.",
    envvar=f"{PREFIX}_GORDO_SERVER_THREADS",
)
@click.option(
    "--gordo-server-probe-timeout",
    type=int,
    help="timeoutSeconds value for livenessProbe and readinessProbe of Gordo server Deployment",
    envvar=f"{PREFIX}_GORDO_SERVER_PROBE_TIMEOUT",
)
@click.option(
    "--without-prometheus",
    is_flag=True,
    help="Do not deploy Prometheus for Gordo servers monitoring",
    envvar=f"{PREFIX}_WITHOUT_PROMETHEUS",
)
@click.option(
    "--prometheus-metrics-server-workers",
    help="Number of workers for Prometheus metrics servers",
    envvar=f"{PREFIX}_PROMETHEUS_METRICS_SERVER_WORKERS",
    default=1,
)
@click.option(
    "--image-pull-policy",
    help="Default imagePullPolicy for all gordo's images",
    envvar=f"{PREFIX}_IMAGE_PULL_POLICY",
)
@click.option(
    "--with-keda",
    is_flag=True,
    help="Enable support for the KEDA autoscaler",
    envvar=f"{PREFIX}_WITH_KEDA",
)
@click.option(
    "--ml-server-hpa-type",
    help="HPA type for the ML server",
    envvar=f"{PREFIX}_ML_SERVER_HPA_TYPE",
    type=click.Choice(ML_SERVER_HPA_TYPES),
    default=DEFAULT_ML_SERVER_HPA_TYPE,
)
@click.option(
    "--custom-model-builder-envs",
    help="List of custom environment variables in ",
    envvar=f"{PREFIX}_CUSTOM_MODEL_BUILDER_ENVS",
    default=DEFAULT_CUSTOM_MODEL_BUILDER_ENVS,
)
@click.option(
    "--prometheus-server-address",
    help='Prometheus url. Required for "--ml-server-hpa-type=keda"',
    envvar=f"{PREFIX}_PROMETHEUS_SERVER_ADDRESS",
)
@click.option(
    "--keda-prometheus-metric-name",
    help="metricName value for the KEDA prometheus scaler",
    envvar=f"{PREFIX}_KEDA_PROMETHEUS_METRIC_NAME",
    default=DEFAULT_KEDA_PROMETHEUS_METRIC_NAME,
)
@click.option(
    "--keda-prometheus-query",
    help="query value for the KEDA prometheus scaler",
    envvar=f"{PREFIX}_KEDA_PROMETHEUS_QUERY",
    default=DEFAULT_KEDA_PROMETHEUS_QUERY,
)
@click.option(
    "--keda-prometheus-threshold",
    help="threshold value for the KEDA prometheus scaler",
    envvar=f"{PREFIX}_KEDA_PROMETHEUS_THRESHOLD",
    default=DEFAULT_KEDA_PROMETHEUS_THRESHOLD,
)
@click.option(
    "--resources-labels",
    help="Additional labels for resources. Have to be empty string or a dictionary in JSON format",
    envvar=f"{PREFIX}_RESOURCE_LABELS",
    default="",
)
@click.pass_context
def workflow_generator_cli(gordo_ctx, **ctx):
    """
    Machine Configuration to Argo Workflow
    """

    context: Dict[Any, Any] = ctx.copy()
    yaml_content = wg.get_dict_from_yaml(context["machine_config"])

    try:
        log_level = yaml_content["globals"]["runtime"]["log_level"]
    except KeyError:
        log_level = os.getenv("GORDO_LOG_LEVEL", gordo_ctx.obj["log_level"])

    logging.getLogger("gordo").setLevel(log_level.upper())
    context["log_level"] = log_level.upper()

    validate_generate_context(context)

    context["resources_labels"] = prepare_resources_labels(context["resources_labels"])

    model_builder_env = None
    if context["custom_model_builder_envs"]:
        model_builder_env = json.loads(context["custom_model_builder_envs"])
    # Create normalized config
    config = NormalizedConfig(
        yaml_content,
        project_name=context["project_name"],
        model_builder_env=model_builder_env,
    )

    version = parse_version(context["gordo_version"])
    if "image_pull_policy" not in context or not context["image_pull_policy"]:
        context["image_pull_policy"] = wg.default_image_pull_policy(version)
    logger.info(
        "Generate config with gordo_version=%s and imagePullPolicy=%s",
        context["gordo_version"],
        context["image_pull_policy"],
    )

    context["max_server_replicas"] = (
        context.pop("n_servers") or len(config.machines) * 10
    )

    context["volumes"] = None
    if "volumes" in config.globals["runtime"]:
        context["volumes"] = config.globals["runtime"]["volumes"]

    builder_runtime = config.globals["runtime"]["builder"]
    # We know these exist since we set them in the default globals
    builder_resources = builder_runtime["resources"]
    context["model_builder_resources_requests_memory"] = builder_resources["requests"][
        "memory"
    ]
    context["model_builder_resources_requests_cpu"] = builder_resources["requests"][
        "cpu"
    ]
    context["model_builder_resources_limits_memory"] = builder_resources["limits"][
        "memory"
    ]
    context["model_builder_resources_limits_cpu"] = builder_resources["limits"]["cpu"]
    context["model_builder_image"] = config.globals["runtime"]["builder"]["image"]

    context["builder_runtime"] = builder_runtime

    context["server_resources"] = config.globals["runtime"]["server"]["resources"]
    context["server_image"] = config.globals["runtime"]["server"]["image"]

    context["prometheus_metrics_server_resources"] = config.globals["runtime"][
        "prometheus_metrics_server"
    ]["resources"]

    context["prometheus_metrics_server_image"] = config.globals["runtime"][
        "prometheus_metrics_server"
    ]["image"]

    context["deployer_image"] = config.globals["runtime"]["deployer"]["image"]

    # These are also set in the default globals, and guaranteed to exist
    client_resources = config.globals["runtime"]["client"]["resources"]
    context["client_resources_requests_memory"] = client_resources["requests"]["memory"]
    context["client_resources_requests_cpu"] = client_resources["requests"]["cpu"]
    context["client_resources_limits_memory"] = client_resources["limits"]["memory"]
    context["client_resources_limits_cpu"] = client_resources["limits"]["cpu"]

    context["client_image"] = config.globals["runtime"]["client"]["image"]

    context["client_max_instances"] = config.globals["runtime"]["client"][
        "max_instances"
    ]

    influx_resources = config.globals["runtime"]["influx"]["resources"]
    context["influx_resources_requests_memory"] = influx_resources["requests"]["memory"]
    context["influx_resources_requests_cpu"] = influx_resources["requests"]["cpu"]
    context["influx_resources_limits_memory"] = influx_resources["limits"]["memory"]
    context["influx_resources_limits_cpu"] = influx_resources["limits"]["cpu"]

    nr_of_models_with_clients = len(
        [
            machine
            for machine in config.machines
            if machine.runtime.get("influx", {}).get("enable", True)
        ]
    )
    context["client_total_instances"] = nr_of_models_with_clients

    # Should we start up influx/grafana at all, i.e. is there at least one request
    # for it?"
    enable_influx = nr_of_models_with_clients > 0
    context["enable_influx"] = enable_influx

    context["postgres_host"] = f"gordo-postgres-{config.project_name}"

    context["keda_prometheus_query"] = prepare_keda_prometheus_query(context)

    # If enabling influx, we setup a postgres reporter to send metadata
    # to allowing querying about the machine from grafana
    if enable_influx:
        pg_reporter = {
            "gordo.reporters.postgres.PostgresReporter": {
                "host": context["postgres_host"]
            }
        }
        for machine in config.machines:
            machine.runtime["reporters"].append(pg_reporter)

    # Determine if MlFlowReporter should be enabled per machine
    for machine in config.machines:
        try:
            enabled = machine.runtime["builder"]["remote_logging"]["enable"]
        except KeyError:
            continue
        else:
            if enabled:
                machine.runtime["reporters"].append(
                    "gordo.reporters.mlflow.MlFlowReporter"
                )

    context["machines"] = config.machines

    # Context requiring pre-processing
    context["target_names"] = [machine.name for machine in config.machines]

    # Json dump owner_references, if not None, otherwise pop it out of the context
    if context["owner_references"]:
        context["owner_references"] = json.dumps(context["owner_references"])
    else:
        context.pop("owner_references")

    builder_exceptions_report_level = get_builder_exceptions_report_level(config)
    context["builder_exceptions_report_level"] = builder_exceptions_report_level.name
    if builder_exceptions_report_level != ReportLevel.EXIT_CODE:
        context["builder_exceptions_report_file"] = "/tmp/exception.json"

    if context["workflow_template"]:
        template = wg.load_workflow_template(context["workflow_template"])
    else:
        workflow_template = pkg_resources.resource_filename(
            "gordo.workflow.workflow_generator.resources", "argo-workflow.yml.template"
        )
        template = wg.load_workflow_template(workflow_template)

    # Clear output file
    if context["output_file"]:
        open(context["output_file"], "w").close()  # type: ignore
    project_workflow = 0
    for i in range(0, len(config.machines), context["split_workflows"]):  # type: ignore
        logger.info(
            f"Generating workflow for machines {i} to {i + context['split_workflows']}"
        )
        context["machines"] = config.machines[i : i + context["split_workflows"]]
        context["project_workflow"] = str(project_workflow)

        if context["output_file"]:
            s = template.stream(**context)
            with open(context["output_file"], "a") as f:  # type: ignore
                if i != 0:
                    f.write("\n---\n")
                s.dump(f)
        else:
            output = template.render(**context)
            if i != 0:
                print("\n---\n")
            print(output)
        project_workflow += 1


@click.command("unique-tags")
@click.option(
    "--machine-config", type=str, required=True, help="Machine configuration file"
)
@click.option(
    "--output-file-tag-list",
    type=str,
    required=False,
    help="Optional file to dump list of unique tags",
)
def unique_tag_list_cli(machine_config: str, output_file_tag_list: str):

    yaml_content = wg.get_dict_from_yaml(machine_config)

    machines = NormalizedConfig(yaml_content, project_name="test-proj-name").machines

    tag_list = set(tag for machine in machines for tag in machine.dataset.tag_list)

    if output_file_tag_list:
        with open(output_file_tag_list, "w") as output_file:
            for tag in tag_list:
                output_file.write(f"{tag.name}\n")
    else:
        for tag in tag_list:
            print(tag.name)


workflow_cli.add_command(workflow_generator_cli)
workflow_cli.add_command(unique_tag_list_cli)

if __name__ == "__main__":
    workflow_cli()



import timeit

from copy import copy
from flask import Flask, g, request, Request, Response
from typing import Optional, Tuple, List, Dict, Iterable
from prometheus_client.multiprocess import MultiProcessCollector
from prometheus_client.registry import CollectorRegistry
from prometheus_client import Counter, Histogram, Gauge
from http import HTTPStatus


def create_registry():
    registry = CollectorRegistry()
    MultiProcessCollector(registry)
    return registry


def to_status_code(response_status):
    if isinstance(response_status, HTTPStatus):
        return response_status.value
    else:
        return response_status


def url_rule_to_str(url_rule):
    return url_rule.rule


def current_time():
    return timeit.default_timer()


class GordoServerPrometheusMetrics:
    """
    Container for encapsulating all Prometheus related logic
    The simplest way to use with preexisting Flask application:

    >>> from flask import Flask
    >>> from prometheus_client.registry import CollectorRegistry
    >>> app = Flask("test")
    >>> @app.route('/hello')
    ... def hello():
    ...     return 'Hello, World'
    >>> prometheus_metrics = GordoServerPrometheusMetrics(registry=CollectorRegistry())
    >>> prometheus_metrics.prepare_app(app)
    """

    prefix = "gordo_server"
    main_labels = ("method", "path", "status_code")

    @staticmethod
    def main_label_values(req: Request, resp: Response):
        return (
            req.method,
            url_rule_to_str(req.url_rule),
            to_status_code(resp.status_code),
        )

    def __init__(
        self,
        args_labels: Optional[Iterable[Tuple[str, str]]] = None,
        info: Optional[Dict[str, str]] = None,
        ignore_paths: Optional[Iterable[str]] = None,
        registry: Optional[CollectorRegistry] = None,
    ):
        self.args_labels = args_labels if args_labels is not None else []
        if ignore_paths is not None:
            ignore_paths = set(ignore_paths)
        self.ignore_paths = ignore_paths if ignore_paths is not None else {}
        self.info = info
        self.label_names: List[str] = []
        self.label_values: List[str] = []
        self.args_names: List[str] = []

        if registry is None:
            registry = create_registry()
        self.registry = registry
        self.init_labels()
        self.request_duration_seconds = Histogram(
            "%s_request_duration_seconds" % self.prefix,
            "HTTP request duration, in seconds",
            self.label_names,
            registry=registry,
        )
        self.request_count = Counter(
            "%s_requests_total" % self.prefix,
            "Total HTTP requests",
            self.label_names,
            registry=registry,
        )

    def init_labels(self):
        label_names, label_values = [], []
        if self.info is not None:
            for name, value in self.info.items():
                label_names.append(name)
                label_values.append(value)
            gauge_info = Gauge(
                self.prefix + "_info",
                "Gordo information",
                label_names,
                registry=self.registry,
            )
            gauge_info = gauge_info.labels(*label_values)
            gauge_info.set(1)
        args_names = []
        for arg_name, label_name in self.args_labels:
            args_names.append(arg_name)
            label_names.append(label_name)
        self.args_names = args_names
        label_names.extend(self.main_labels)
        self.label_names = label_names
        self.label_values = label_values

    def request_label_values(self, req: Request, resp: Response):
        label_values = copy(self.label_values)
        view_args = req.view_args
        for arg_name in self.args_names:
            value = view_args.get(arg_name, "") if view_args is not None else ""
            label_values.append(value)
        label_values.extend(self.main_label_values(req, resp))
        return label_values

    def prepare_app(self, app: Flask):
        @app.before_request
        def _start_prometheus():
            g.prometheus_metrics = self
            g.prometheus_start_time = current_time()

        @app.after_request
        def _end_prometheus(response: Response) -> Response:
            url_rule = url_rule_to_str(request.url_rule)
            if self.ignore_paths is not None and url_rule in self.ignore_paths:
                return response
            label_values = self.request_label_values(request, response)
            self.request_duration_seconds.labels(*label_values).observe(
                current_time() - g.prometheus_start_time
            )
            self.request_count.labels(*label_values).inc(1)
            del g.prometheus_metrics
            return response



import abc
import pandas as pd
import xarray as xr
from typing import Optional, Union
from datetime import timedelta
from sklearn.base import BaseEstimator

from gordo.machine.model.base import GordoBase


class AnomalyDetectorBase(BaseEstimator, GordoBase, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def anomaly(
        self,
        X: Union[pd.DataFrame, xr.DataArray],
        y: Union[pd.DataFrame, xr.DataArray],
        frequency: Optional[timedelta] = None,
    ) -> Union[pd.DataFrame, xr.Dataset]:
        """
        Take X, y and optionally frequency; returning a dataframe containing
        anomaly score(s)
        """
        ...



# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from gordo.machine.model import utils as model_utils
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error


def test_metrics_wrapper():
    # make the features in y be in different scales
    y = np.array([[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]) * [1, 100]

    # With no scaler provided it is relevant which of the two series gets an 80% error
    metric_func_noscaler = model_utils.metric_wrapper(mean_squared_error)

    mse_feature_one_wrong = metric_func_noscaler(y, y * [0.8, 1])
    mse_feature_two_wrong = metric_func_noscaler(y, y * [1, 0.8])

    assert not np.isclose(mse_feature_one_wrong, mse_feature_two_wrong)

    # With a scaler provided it is not relevant which of the two series gets an 80%
    # error
    scaler = MinMaxScaler().fit(y)
    metric_func_scaler = model_utils.metric_wrapper(mean_squared_error, scaler=scaler)

    mse_feature_one_wrong = metric_func_scaler(y, y * [0.8, 1])
    mse_feature_two_wrong = metric_func_scaler(y, y * [1, 0.8])

    assert np.isclose(mse_feature_one_wrong, mse_feature_two_wrong)


@pytest.mark.parametrize(
    "dates", [pd.date_range("2016-01-01", "2016-01-02", periods=10), None]
)
@pytest.mark.parametrize("tags", [["tag1", "tag2"], ["tag"]])
@pytest.mark.parametrize(
    "target_tag_list",
    (
        ["tag1", "tag2"],
        ["tag3", "tag4"],
        ["tagA"],
        ["tag1"],
        ["tagA", "tagB", "tagC"],
        None,
    ),
)
@pytest.mark.parametrize("output_offset", (0, 1, 2, 3))
def test_base_dataframe_creation(dates, tags, target_tag_list, output_offset):

    # Make model input based on tags
    size = len(dates if dates is not None else list(range(10))) * len(tags)
    model_input = np.random.random(size=size).reshape(-1, len(tags))

    # Model output based on target_tag_list
    size = len(dates if dates is not None else list(range(10))) * len(
        target_tag_list or list(range(20))
    )
    model_output = np.random.random(size=size).reshape((len(model_input), -1))

    # simulate where model's output length doesn't match it's input length
    # ie. as with an LSTM which outputs the offset of it's lookback window
    model_output = model_output[output_offset:]

    # pass in the arrays, of which model output's may be different lengths / shapes than the input
    # but should provide a valid dataframe in all cases.
    df = model_utils.make_base_dataframe(
        tags=tags,
        model_input=model_input,
        model_output=model_output,
        target_tag_list=target_tag_list,
        index=dates,
    )

    # offset column's like 'original-input' since it will be offsetted inside make_base_dataframe()
    assert np.array_equal(df["model-input"].values, model_input[-len(df) :, :])

    # Model input should always have column labels equal to the tags given
    assert df["model-input"].columns.tolist() == tags

    # Ensure model output matches
    assert np.array_equal(df["model-output"].values, model_output[-len(df) :, :])

    # Expected second level column names:
    # If target tags are defined, those should be the names
    if target_tag_list is not None:
        assert target_tag_list == df["model-output"].columns.tolist()

    # If they aren't defined, but model output shape matches input shape, tags should be the names.
    elif model_output.shape[1] == len(tags):
        assert tags == df["model-output"].columns.tolist()

    # Otherwise, column names should be simple range of feature length.
    else:
        assert (
            list(map(str, range(model_output.shape[1])))
            == df["model-output"].columns.tolist()
        )

    # Test expected index if dates were supplied or not
    if dates is not None:
        assert np.array_equal(df.index.values, dates.values[output_offset:])
    else:
        assert np.array_equal(df.index.values, np.arange(0, len(df)))



# -*- coding: utf-8 -*-

from typing import Tuple, Union, Dict, Any

import tensorflow
from tensorflow import keras
from tensorflow.keras.optimizers import Optimizer
from tensorflow.keras.layers import Dense, LSTM
from tensorflow.keras.models import Sequential as KerasSequential

from gordo.machine.model.register import register_model_builder
from gordo.machine.model.factories.utils import hourglass_calc_dims, check_dim_func_len


@register_model_builder(type="KerasLSTMAutoEncoder")
@register_model_builder(type="KerasLSTMForecast")
def lstm_model(
    n_features: int,
    n_features_out: int = None,
    lookback_window: int = 1,
    encoding_dim: Tuple[int, ...] = (256, 128, 64),
    encoding_func: Tuple[str, ...] = ("tanh", "tanh", "tanh"),
    decoding_dim: Tuple[int, ...] = (64, 128, 256),
    decoding_func: Tuple[str, ...] = ("tanh", "tanh", "tanh"),
    out_func: str = "linear",
    optimizer: Union[str, Optimizer] = "Adam",
    optimizer_kwargs: Dict[str, Any] = dict(),
    compile_kwargs: Dict[str, Any] = dict(),
    **kwargs,
) -> tensorflow.keras.models.Sequential:
    """
    Builds a customized Keras LSTM neural network auto-encoder based on a config dict.

    Parameters
    ----------
    n_features: int
        Number of features the dataset X will contain.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.
    lookback_window: int
        Number of timesteps used to train the model.
        One timestep = current observation in the sample.
        Two timesteps = current observation + previous observation in the sample.
        ...
    encoding_dim: tuple
        Tuple of numbers with the number of neurons in the encoding part.
    decoding_dim: tuple
        Tuple of numbers with the number of neurons in the decoding part.
    encoding_func: tuple
        Activation functions for the encoder part.
    decoding_func: tuple
        Activation functions for the decoder part.
    out_func: str
        Activation function for the output Dense layer.
    optimizer: Union[str, Optimizer]
        If str then the name of the optimizer must be provided (e.x. "Adam").
        The arguments of the optimizer can be supplied in optimize_kwargs.
        If a Keras optimizer call the instance of the respective
        class (e.x. Adam(lr=0.01,beta_1=0.9, beta_2=0.999)).  If no arguments are
        provided Keras default values will be set.
    optimizer_kwargs: Dict[str, Any]
        The arguments for the chosen optimizer. If not provided Keras'
        default values will be used.
    compile_kwargs: Dict[str, Any]
        Parameters to pass to ``keras.Model.compile``.

    Returns
    -------
    keras.models.Sequential
        Returns Keras sequential model.

    """
    n_features_out = n_features_out or n_features

    check_dim_func_len("encoding", encoding_dim, encoding_func)
    check_dim_func_len("decoding", decoding_dim, decoding_func)

    model = KerasSequential()

    # encoding layers
    kwargs = {"return_sequences": True}
    for i, (n_neurons, activation) in enumerate(zip(encoding_dim, encoding_func)):
        input_shape = (lookback_window, n_neurons if i != 0 else n_features)
        kwargs.update(dict(activation=activation, input_shape=input_shape))
        model.add(LSTM(n_neurons, **kwargs))

    # decoding layers
    for i, (n_neurons, activation) in enumerate(zip(decoding_dim, decoding_func)):
        return_seq = i != len(decoding_dim) - 1  # Don't return sequences on 2nd to last
        model.add(LSTM(n_neurons, activation=activation, return_sequences=return_seq))

    # output layer
    if isinstance(optimizer, str):
        Optim = getattr(keras.optimizers, optimizer)
        optimizer = Optim(**optimizer_kwargs)

    model.add(Dense(units=n_features_out, activation=out_func))

    # Update kwargs and compile model
    compile_kwargs.setdefault("loss", "mse")
    compile_kwargs.update({"optimizer": optimizer})
    model.compile(**compile_kwargs)
    return model


@register_model_builder(type="KerasLSTMAutoEncoder")
@register_model_builder(type="KerasLSTMForecast")
def lstm_symmetric(
    n_features: int,
    n_features_out: int = None,
    lookback_window: int = 1,
    dims: Tuple[int, ...] = (256, 128, 64),
    funcs: Tuple[str, ...] = ("tanh", "tanh", "tanh"),
    out_func: str = "linear",
    optimizer: Union[str, Optimizer] = "Adam",
    optimizer_kwargs: Dict[str, Any] = dict(),
    compile_kwargs: Dict[str, Any] = dict(),
    **kwargs,
) -> tensorflow.keras.models.Sequential:
    """
    Builds a symmetrical lstm model

    Parameters
    ----------
    n_features: int
         Number of input and output neurons.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.
    lookback_window: int
        Number of timesteps used to train the model.
        One timestep = sample contains current observation.
        Two timesteps = sample contains current and previous observation.
        ...
    dims: Tuple[int,...]
         Number of neurons per layers for the encoder, reversed for the decoder.
         Must have len > 0
    funcs: List[str]
        Activation functions for the internal layers.
    out_func: str
        Activation function for the output Dense layer.
    optimizer: Union[str, Optimizer]
        If str then the name of the optimizer must be provided (e.x. "Adam").
        The arguments of the optimizer can be supplied in optimization_kwargs.
        If a Keras optimizer call the instance of the respective
        class (e.x. Adam(lr=0.01,beta_1=0.9, beta_2=0.999)).  If no arguments are
        provided Keras default values will be set.
    optimizer_kwargs: Dict[str, Any]
        The arguments for the chosen optimizer. If not provided Keras'
        default values will be used.
    compile_kwargs: Dict[str, Any]
        Parameters to pass to ``keras.Model.compile``.

    Returns
    -------
    keras.models.Sequential
        Returns Keras sequential model.
    """

    if len(dims) == 0:
        raise ValueError("Parameter dims must have len > 0")

    return lstm_model(
        n_features=n_features,
        n_features_out=n_features_out,
        lookback_window=lookback_window,
        encoding_dim=dims,
        decoding_dim=dims[::-1],
        encoding_func=funcs,
        decoding_func=funcs[::-1],
        out_func=out_func,
        optimizer=optimizer,
        optimizer_kwargs=optimizer_kwargs,
        compile_kwargs=compile_kwargs,
        **kwargs,
    )


@register_model_builder(type="KerasLSTMAutoEncoder")
@register_model_builder(type="KerasLSTMForecast")
def lstm_hourglass(
    n_features: int,
    n_features_out: int = None,
    lookback_window: int = 1,
    encoding_layers: int = 3,
    compression_factor: float = 0.5,
    func: str = "tanh",
    out_func: str = "linear",
    optimizer: Union[str, Optimizer] = "Adam",
    optimizer_kwargs: Dict[str, Any] = dict(),
    compile_kwargs: Dict[str, Any] = dict(),
    **kwargs,
) -> tensorflow.keras.models.Sequential:

    """

    Builds an hourglass shaped neural network, with decreasing number of neurons
    as one gets deeper into the encoder network and increasing number
    of neurons as one gets out of the decoder network.


    Parameters
    ----------
    n_features: int
        Number of input and output neurons.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.
    encoding_layers: int
        Number of layers from the input layer (exclusive) to the
        narrowest layer (inclusive). Must be > 0. The total nr of layers
        including input and output layer will be 2*encoding_layers + 1.
     compression_factor: float
        How small the smallest layer is as a ratio of n_features
        (smallest layer is rounded up to nearest integer). Must satisfy
        0 <= compression_factor <= 1.
    func: str
        Activation function for the internal layers.
    out_func: str
        Activation function for the output Dense layer.
    optimizer: Union[str, Optimizer]
        If str then the name of the optimizer must be provided (e.x. "Adam").
        The arguments of the optimizer can be supplied in optimization_kwargs.
        If a Keras optimizer call the instance of the respective
        class (e.x. Adam(lr=0.01,beta_1=0.9, beta_2=0.999)).  If no arguments are
        provided Keras default values will be set.
    optimizer_kwargs: Dict[str, Any]
        The arguments for the chosen optimizer. If not provided Keras'
        default values will be used.
    compile_kwargs: Dict[str, Any]
        Parameters to pass to ``keras.Model.compile``.


    Returns
    -------
    keras.models.Sequential

    Examples
    --------
    >>> model = lstm_hourglass(10)
    >>> len(model.layers)
    7
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [8, 7, 5, 5, 7, 8, 10]
    >>> model = lstm_hourglass(5)
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [4, 4, 3, 3, 4, 4, 5]
    >>> model = lstm_hourglass(10, compression_factor=0.2)
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [7, 5, 2, 2, 5, 7, 10]
    >>> model = lstm_hourglass(10, encoding_layers=1)
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [5, 5, 10]
    """
    dims = hourglass_calc_dims(compression_factor, encoding_layers, n_features)

    return lstm_symmetric(
        n_features=n_features,
        n_features_out=n_features_out,
        lookback_window=lookback_window,
        dims=dims,
        funcs=[func] * len(dims),
        out_func=out_func,
        optimizer=optimizer,
        optimizer_kwargs=optimizer_kwargs,
        compile_kwargs=compile_kwargs,
        **kwargs,
    )



# -*- coding: utf-8 -*-
import ipaddress
import typing

import click


class HostIP(click.ParamType):
    """
    Validate input is a valid IP address
    """

    name = "host"

    def convert(self, value, param, ctx):
        try:
            ipaddress.ip_address(value)
            return value
        except ValueError as e:
            self.fail(e)


def key_value_par(val) -> typing.Tuple[str, str]:
    """
    Helpder for CLI input of 'key,val'
    """
    return val.split(",")



from typing import Optional, Any, List, Dict
from pydantic import BaseModel


class ResourceRequirements(BaseModel):
    requests: Optional[Dict[str, Any]]
    limits: Optional[Dict[str, Any]]


class ConfigMapKeySelector(BaseModel):
    key: str
    value: str


class SecretKeySelector(BaseModel):
    key: str
    name: str


class EnvVarSource(BaseModel):
    configMapKeyRef: Optional[ConfigMapKeySelector]
    secretKeyRef: Optional[SecretKeySelector]


class EnvVar(BaseModel):
    name: str
    value: Optional[str]
    valueFrom: Optional[EnvVarSource]


class ObjectMeta(BaseModel):
    labels: Optional[Dict[str, Any]]


class CSIVolumeSource(BaseModel):
    driver: str
    readOnly: bool = False
    fsType: Optional[str]
    volumeAttributes: Optional[Dict[str, Any]]


class Volume(BaseModel):
    name: str
    csi: Optional[CSIVolumeSource]


class VolumeMount(BaseModel):
    name: str
    readOnly: Optional[bool]
    mountPath: str


class PodRuntime(BaseModel):
    image: str
    resources: Optional[ResourceRequirements]
    metadata: Optional[ObjectMeta]
    env: Optional[List[EnvVar]]
    volumeMounts: Optional[List[VolumeMount]]


class RemoteLogging(BaseModel):
    enable: bool = False


class BuilderPodRuntime(PodRuntime):
    remote_logging: RemoteLogging



# -*- coding: utf-8 -*-

"""
Functions to be used within sklearn's FunctionTransformer
https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.FunctionTransformer.html

Each function SHALL take an X, and optionally a y.

Functions CAN take additional arguments which should be given during the initialization of the FunctionTransformer

Example:

>>> from sklearn.preprocessing import FunctionTransformer
>>> import numpy as np
>>> def my_function(X, another_arg):
...     # Some fancy X manipulation...
...     return X
>>> transformer = FunctionTransformer(func=my_function, kw_args={'another_arg': 'this thing'})
>>> out = transformer.fit_transform(np.random.random(100).reshape(10, 10))
"""


def multiply_by(X, factor):
    """
    Multiplies X by a given factor
    """
    return X * factor



# -*- coding: utf-8 -*-

import unittest
from unittest import mock

from gordo.machine.model.factories.feedforward_autoencoder import (
    feedforward_symmetric,
    feedforward_hourglass,
)


def feedforward_model_mocker(
    n_features: int,
    n_features_out: int,
    encoding_dim,
    decoding_dim,
    encoding_func,
    decoding_func,
    optimizer,
    optimizer_kwargs,
    compile_kwargs,
):
    return (
        n_features,
        n_features_out,
        encoding_dim,
        decoding_dim,
        encoding_func,
        decoding_func,
        optimizer,
        optimizer_kwargs,
        compile_kwargs,
    )


class FeedForwardAutoEncoderTestCase(unittest.TestCase):
    @mock.patch(
        "gordo.machine.model.factories.feedforward_autoencoder.feedforward_model",
        side_effect=feedforward_model_mocker,
    )
    def test_feedforward_symmetric_basic(self, _):
        """
        Test that feedforward_symmetric calls feedforward_model correctly
        """
        (
            n_features,
            n_features_out,
            encoding_dim,
            decoding_dim,
            encoding_func,
            decoding_func,
            optimizer,
            optimizer_kwargs,
            compile_kwargs,
        ) = feedforward_symmetric(
            5, 5, (4, 3, 2, 1), ("relu", "relu", "tanh", "tanh"), {}
        )
        self.assertEqual(n_features, 5)
        self.assertEqual(compile_kwargs, {})
        self.assertEqual(encoding_dim, (4, 3, 2, 1))
        self.assertEqual(decoding_dim, (1, 2, 3, 4))
        self.assertEqual(encoding_func, ("relu", "relu", "tanh", "tanh"))
        self.assertEqual(decoding_func, ("tanh", "tanh", "relu", "relu"))

    def test_feedforward_symmetric_checks_dims(self):
        """
        Test that feedforward_symmetric validates parameter requirements
        """
        with self.assertRaises(ValueError):
            feedforward_symmetric(4, (), ())

    @mock.patch(
        "gordo.machine.model.factories.feedforward_autoencoder.feedforward_model",
        side_effect=feedforward_model_mocker,
    )
    def test_feedforward_hourglass_basic(self, _):
        """
        Test that feedforward_hourglass calls feedforward_model correctly
        """
        (
            n_features,
            n_features_out,
            encoding_dim,
            decoding_dim,
            encoding_func,
            decoding_func,
            optimizer,
            optimizer_kwargs,
            compile_kwargs,
        ) = feedforward_hourglass(10, 10, func="relu")
        self.assertEqual(n_features, 10)
        self.assertEqual(n_features_out, 10)
        self.assertEqual(encoding_dim, (8, 7, 5))
        self.assertEqual(decoding_dim, (5, 7, 8))
        self.assertEqual(encoding_func, ("relu", "relu", "relu"))
        self.assertEqual(decoding_func, ("relu", "relu", "relu"))

        (
            n_features,
            n_features_out,
            encoding_dim,
            decoding_dim,
            encoding_func,
            decoding_func,
            optimizer,
            optimizer_kwargs,
            compile_kwargs,
        ) = feedforward_hourglass(3, 3)
        self.assertEqual(n_features, 3)
        self.assertEqual(n_features_out, 3)
        self.assertEqual(encoding_dim, (3, 2, 2))
        self.assertEqual(decoding_dim, (2, 2, 3))
        self.assertEqual(encoding_func, ("tanh", "tanh", "tanh"))
        self.assertEqual(decoding_func, ("tanh", "tanh", "tanh"))

        (
            n_features,
            n_features_out,
            encoding_dim,
            decoding_dim,
            encoding_func,
            decoding_func,
            optimizer,
            optimizer_kwargs,
            compile_kwargs,
        ) = feedforward_hourglass(10, 10, compression_factor=0.3)
        self.assertEqual(n_features, 10)
        self.assertEqual(n_features_out, 10)
        self.assertEqual(encoding_dim, (8, 5, 3))
        self.assertEqual(decoding_dim, (3, 5, 8))
        self.assertEqual(encoding_func, ("tanh", "tanh", "tanh"))
        self.assertEqual(decoding_func, ("tanh", "tanh", "tanh"))

    @mock.patch(
        "gordo.machine.model.factories.feedforward_autoencoder.feedforward_model",
        side_effect=feedforward_model_mocker,
    )
    def test_feedforward_hourglass_compression_factors(self, _):
        """
        Test that feedforward_hourglass handles compression_factor=1 and
        compression_factor=0 correctly.
        """

        # compression_factor=1 is no compression
        (
            n_features,
            n_features_out,
            encoding_dim,
            decoding_dim,
            encoding_func,
            decoding_func,
            optimizer,
            optimizer_kwargs,
            compile_kwargs,
        ) = feedforward_hourglass(10, 10, compression_factor=1)
        self.assertEqual(n_features, 10)
        self.assertEqual(n_features_out, 10)
        self.assertEqual(encoding_dim, (10, 10, 10))
        self.assertEqual(decoding_dim, (10, 10, 10))
        self.assertEqual(encoding_func, ("tanh", "tanh", "tanh"))
        self.assertEqual(decoding_func, ("tanh", "tanh", "tanh"))

        # compression_factor=0 has smallest layer with 1 node.
        (
            n_features,
            n_features_out,
            encoding_dim,
            decoding_dim,
            encoding_func,
            decoding_func,
            optimizer,
            optimizer_kwargs,
            compile_kwargs,
        ) = feedforward_hourglass(100000, 100000, compression_factor=0)
        self.assertEqual(n_features, 100000)
        self.assertEqual(n_features_out, 100000)
        self.assertEqual(encoding_dim, (66667, 33334, 1))
        self.assertEqual(decoding_dim, (1, 33334, 66667))
        self.assertEqual(encoding_func, ("tanh", "tanh", "tanh"))
        self.assertEqual(decoding_func, ("tanh", "tanh", "tanh"))

    def test_feedforward_hourglass_checks_enc_layers(self):
        """
        Test that feedforward_hourglass validates parameter requirements
        """
        with self.assertRaises(ValueError):
            feedforward_hourglass(3, encoding_layers=0)

    def test_feedforward_hourglass_checks_compression_factor(self):
        """
        Test that feedforward_hourglass validates parameter requirements
        """
        with self.assertRaises(ValueError):
            feedforward_hourglass(3, compression_factor=2)
        with self.assertRaises(ValueError):
            feedforward_hourglass(3, compression_factor=-1)



from flask import Flask
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from prometheus_client.exposition import make_wsgi_app
from .metrics import create_registry


def build_app() -> Flask:
    """
    Creating Flask application in order to serve all Prometheus metrics

    :return: Flask
    """
    curr_app = Flask("gordoserver_prometheus")

    registry = create_registry()

    curr_app.wsgi_app = DispatcherMiddleware(  # type: ignore
        curr_app.wsgi_app, {"/metrics": make_wsgi_app(registry)}
    )

    @curr_app.route("/healthcheck")
    def health_check():
        return "", 200

    return curr_app



# -*- coding: utf-8 -*-

import abc
import logging
import io
import importlib
from pprint import pformat
from typing import Union, Callable, Dict, Any, Optional, Tuple
from abc import ABCMeta
from copy import copy, deepcopy
from importlib.util import find_spec

import h5py
import tensorflow.keras.models
from tensorflow.keras.models import load_model, save_model
from tensorflow.keras.preprocessing.sequence import pad_sequences, TimeseriesGenerator
from tensorflow.keras.wrappers.scikit_learn import KerasRegressor as BaseWrapper
from tensorflow.keras.callbacks import History
import numpy as np
import pandas as pd
import xarray as xr

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.metrics import explained_variance_score
from sklearn.exceptions import NotFittedError

from gordo import serializer
from gordo.machine.model.base import GordoBase

# This is required to run `register_model_builder` against registered factories
from gordo.machine.model.factories import *  # pragma: no flakes

from gordo.machine.model.register import register_model_builder

logger = logging.getLogger(__name__)


class KerasBaseEstimator(BaseWrapper, GordoBase, BaseEstimator):
    supported_fit_args = [
        "batch_size",
        "epochs",
        "verbose",
        "callbacks",
        "validation_split",
        "shuffle",
        "class_weight",
        "initial_epoch",
        "steps_per_epoch",
        "validation_batch_size",
        "max_queue_size",
        "workers",
        "use_multiprocessing",
    ]

    def __init__(
        self,
        kind: Union[
            str, Callable[[int, Dict[str, Any]], tensorflow.keras.models.Model]
        ],
        **kwargs,
    ) -> None:
        """
        Initialized a Scikit-Learn API compatitble Keras model with a pre-registered
        function or a builder function
        directly.

        Parameters
        ----------
        kind: Union[callable, str]
            The structure of the model to build. As designated by any registered builder
            functions, registered with
            `gordo_compontents.model.register.register_model_builder`.
            Alternatively, one may pass a builder function directly to this argument.
            Such a function should accept `n_features` as it's first argument, and pass
            any additional parameters to `**kwargs`

        kwargs: dict
            Any additional args which are passed to the factory
            building function and/or any additional args to be passed
            to Keras' fit() method
        """
        self.build_fn = None
        self.history = None

        self.kind = self.load_kind(kind)
        self.kwargs: Dict[str, Any] = kwargs

    @staticmethod
    def parse_module_path(module_path) -> Tuple[Optional[str], str]:
        module_paths = module_path.split(".")
        if len(module_paths) == 1:
            return None, module_paths[0]
        else:
            return ".".join(module_paths[:-1]), module_paths[-1]

    def load_kind(self, kind):
        if callable(kind):
            register_model_builder(type=self.__class__.__name__)(kind)
            return kind.__name__
        else:
            module_name, class_name = self.parse_module_path(kind)
            if module_name is None:
                if (
                    class_name
                    not in register_model_builder.factories[self.__class__.__name__]
                ):
                    raise ValueError(
                        f"kind: {kind} is not an available model for type: {class_name}!"
                    )
            else:
                has_error = True
                try:
                    has_error = not find_spec(module_name)
                except ModuleNotFoundError:
                    pass
                if has_error:
                    raise ValueError(
                        f"kind: {kind}, unable to find module: '{module_name}'"
                    )
            return kind

    @classmethod
    def extract_supported_fit_args(cls, kwargs):
        """
        Filtering only ``fit`` related kwargs

        Parameters
        ----------
        kwargs: dict

        Returns
        -------

        """
        fit_args = {}
        for arg in cls.supported_fit_args:
            if arg in kwargs:
                fit_args[arg] = kwargs[arg]
        return fit_args

    @classmethod
    def from_definition(cls, definition: dict):
        """
        Handler for ``gordo.serializer.from_definition``

        Parameters
        ----------
        definition: dict

        Returns
        -------

        """
        kind = definition.pop("kind")
        kwargs = copy(definition)
        return cls(kind, **kwargs)

    def into_definition(self) -> dict:
        """
        Handler for ``gordo.serializer.into_definition``

        Returns
        -------
        dict

        """
        definition = copy(self.kwargs)
        definition["kind"] = self.kind
        return definition

    @property
    def sk_params(self):
        """
        Parameters used for scikit learn kwargs"""
        fit_args = self.extract_supported_fit_args(self.kwargs)
        if fit_args:
            kwargs = deepcopy(self.kwargs)
            kwargs.update(serializer.load_params_from_definition(fit_args))
            return kwargs
        else:
            return self.kwargs

    def __getstate__(self):

        state = self.__dict__.copy()

        if hasattr(self, "model") and self.model is not None:
            buf = io.BytesIO()
            with h5py.File(buf, compression="lzf", mode="w") as h5:
                save_model(self.model, h5, overwrite=True, save_format="h5")
                buf.seek(0)
                state["model"] = buf
            if hasattr(self, "history"):
                from tensorflow.python.keras.callbacks import History

                history = History()
                history.history = self.history.history
                history.params = self.history.params
                history.epoch = self.history.epoch
                state["history"] = history
        return state

    def __setstate__(self, state):
        if "model" in state:
            with h5py.File(state["model"], compression="lzf", mode="r") as h5:
                state["model"] = load_model(h5, compile=False)
        self.__dict__ = state
        return self

    @staticmethod
    def get_n_features_out(
        y: Union[np.ndarray, pd.DataFrame, xr.DataArray]
    ) -> Union[int, tuple]:
        shape_len = len(y.shape)
        if shape_len == 1:
            raise ValueError(
                "Unsupported number of the output dataset dimensions %d" % shape_len
            )
        elif shape_len == 2:
            return y.shape[1]
        else:
            return y.shape[1:]

    @staticmethod
    def get_n_features(
        X: Union[np.ndarray, pd.DataFrame, xr.DataArray]
    ) -> Union[int, tuple]:
        shape_len = len(X.shape)
        if shape_len == 1:
            raise ValueError(
                "Unsupported number of the output dataset dimensions %d" % shape_len
            )
        elif shape_len == 2:
            return X.shape[1]
        else:
            # TODO fix for the legacy LSTM
            if not isinstance(X, xr.DataArray):
                return X.shape[2]
            return X.shape[1:]

    def fit(
        self,
        X: Union[np.ndarray, pd.DataFrame, xr.DataArray],
        y: Union[np.ndarray, pd.DataFrame, xr.DataArray],
        **kwargs,
    ):
        """
        Fit the model to X given y.

        Parameters
        ----------
        X: Union[np.ndarray, pd.DataFrame, xr.Dataset]
            numpy array or pandas dataframe
        y: Union[np.ndarray, pd.DataFrame, xr.Dataset]
            numpy array or pandas dataframe
        sample_weight: np.ndarray
            array like - weight to assign to samples
        kwargs
            Any additional kwargs to supply to keras fit method.

        Returns
        -------
        self
            'KerasAutoEncoder'
        """

        # Reshape y if needed, and set n features of target
        if isinstance(y, np.ndarray) and y.ndim == 1:
            y = y.reshape(-1, 1)

        logger.debug(f"Fitting to data of length: {len(X)}")
        self.kwargs.update(
            {
                "n_features": self.get_n_features(X),
                "n_features_out": self.get_n_features_out(y),
            }
        )

        if isinstance(X, (pd.DataFrame, xr.DataArray)):
            X = X.values
        if isinstance(y, (pd.DataFrame, xr.DataArray)):
            y = y.values
        kwargs.setdefault("verbose", 0)
        history = super().fit(X, y, sample_weight=None, **kwargs)
        if isinstance(history, History):
            self.history = history
        return self

    def predict(self, X: np.ndarray, **kwargs) -> np.ndarray:
        """

        Parameters
        ----------
        X: np.ndarray
            Input data
        kwargs: dict
            kwargs which are passed to Kera's ``predict`` method


        Returns
        -------
        results:
            np.ndarray
        """
        return self.model.predict(X, **kwargs)

    def get_params(self, **params):
        """
        Gets the parameters for this estimator

        Parameters
        ----------
        params
            ignored (exists for API compatibility).

        Returns
        -------
        Dict[str, Any]
            Parameters used in this estimator
        """
        params = super().get_params(**params)
        params.pop("build_fn", None)
        params.update({"kind": self.kind})
        params.update(self.kwargs)
        return params

    def __call__(self):
        module_name, class_name = self.parse_module_path(self.kind)
        if module_name is None:
            factories = register_model_builder.factories[self.__class__.__name__]
            build_fn = factories[self.kind]
        else:
            module = importlib.import_module(module_name)
            if not hasattr(module, class_name):
                raise ValueError(
                    "kind: %s, unable to find class %s in module '%s'"
                    % (self.kind, class_name, module_name)
                )
            build_fn = getattr(module, class_name)
        return build_fn(**self.sk_params)

    def get_metadata(self):
        """
        Get metadata for the KerasBaseEstimator.
        Includes a dictionary with key "history". The key's value is a a dictionary
        with a key "params" pointing another dictionary with various parameters.
        The metrics are defined in the params dictionary under "metrics".
        For each of the metrics there is a key who's value is a list of values for this
        metric per epoch.

        Returns
        -------
        Dict
            Metadata dictionary, including a history object if present
        """
        if hasattr(self, "model") and hasattr(self, "history"):
            history = self.history.history
            history["params"] = self.history.params
            return {"history": history}
        else:
            return {}


class KerasAutoEncoder(KerasBaseEstimator, TransformerMixin):
    """
    Subclass of the KerasBaseEstimator to allow fitting to just X without requiring y.
    """

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        """
        Returns the explained variance score between auto encoder's input vs output

        Parameters
        ----------
        X: Union[np.ndarray, pd.DataFrame]
            Input data to the model
        y: Union[np.ndarray, pd.DataFrame]
            Target
        sample_weight: Optional[np.ndarray]
            sample weights

        Returns
        -------
        score: float
            Returns the explained variance score
        """
        if not hasattr(self, "model"):
            raise NotFittedError(
                f"This {self.__class__.__name__} has not been fitted yet."
            )

        out = self.model.predict(X)

        return explained_variance_score(y, out)


class KerasRawModelRegressor(KerasAutoEncoder):
    """
    Create a scikit-learn like model with an underlying tensorflow.keras model
    from a raw config.
    Examples
    --------
    >>> import yaml
    >>> import numpy as np
    >>> config_str = '''
    ...   # Arguments to the .compile() method
    ...   compile:
    ...     loss: mse
    ...     optimizer: adam
    ...
    ...   # The architecture of the model itself.
    ...   spec:
    ...     tensorflow.keras.models.Sequential:
    ...       layers:
    ...         - tensorflow.keras.layers.Dense:
    ...             units: 4
    ...         - tensorflow.keras.layers.Dense:
    ...             units: 1
    ... '''
    >>> config = yaml.safe_load(config_str)
    >>> model = KerasRawModelRegressor(kind=config)
    >>>
    >>> X, y = np.random.random((10, 4)), np.random.random((10, 1))
    >>> model.fit(X, y, verbose=0)
    KerasRawModelRegressor(kind: {'compile': {'loss': 'mse', 'optimizer': 'adam'},
     'spec': {'tensorflow.keras.models.Sequential': {'layers': [{'tensorflow.keras.layers.Dense': {'units': 4}},
                                                                {'tensorflow.keras.layers.Dense': {'units': 1}}]}}})
    >>> out = model.predict(X)
    """

    _expected_keys = ("spec", "compile")

    def load_kind(self, kind):
        return kind

    def __repr__(self):
        return f"{self.__class__.__name__}(kind: {pformat(self.kind)})"

    def __call__(self):
        """Build Keras model from specification"""
        if not all(k in self.kind for k in self._expected_keys):
            raise ValueError(
                f"Expected spec to have keys: {self._expected_keys}, but found {self.kind.keys()}"
            )
        logger.debug(f"Building model from spec: {self.kind}")

        model = serializer.from_definition(self.kind["spec"])

        # Load any compile kwargs as well, such as compile.optimizer which may map to class obj
        kwargs = serializer.from_definition(self.kind["compile"])

        model.compile(**kwargs)
        return model


class KerasLSTMBaseEstimator(KerasBaseEstimator, TransformerMixin, metaclass=ABCMeta):
    """
    Abstract Base Class to allow to train a many-one LSTM autoencoder and an LSTM
    1 step forecast
    """

    def __init__(
        self,
        kind: Union[Callable, str],
        lookback_window: int = 1,
        batch_size: int = 32,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        kind: Union[Callable, str]
            The structure of the model to build. As designated by any registered builder
            functions, registered with
            `gordo.machine.model.register.register_model_builder`.
            Alternatively, one may pass a builder function directly to this argument.
            Such a function should accept `n_features` as it's first argument, and pass
            any additional parameters to `**kwargs`.
        lookback_window: int
            Number of timestamps (lags) used to train the model.
        batch_size: int
            Number of training examples used in one epoch.
        epochs: int
            Number of epochs to train the model. An epoch is an iteration over the
            entire data provided.
        verbose: int
            Verbosity mode. Possible values are 0, 1, or 2 where 0 = silent,
            1 = progress bar, 2 = one line per epoch.
        kwargs: dict
            Any arguments which are passed to the factory building function and/or any
            additional args to be passed to the intermediate fit method.
        """
        self.lookback_window = lookback_window
        self.batch_size = batch_size
        kwargs["lookback_window"] = lookback_window
        kwargs["kind"] = kind
        kwargs["batch_size"] = batch_size

        # fit_generator_params is a set of strings with the keyword arguments of
        # Keras fit_generator method (excluding "shuffle" as this will be hardcoded).
        # This will be used in the fit method of the respective subclasses to match
        # the kwargs supplied when instantiating the subclass. The matched kwargs
        # will override the default kwargs of Keras fit_generator method when
        # training the model. Note: The decorator
        # "@interfaces.legacy_generator_methods_support" to Keras' fit_generator
        # method does not forward any arguments to the inspect module
        self.fit_generator_params = {
            "steps_per_epoch",
            "epochs",
            "verbose",
            "callbacks",
            "validation_data",
            "validation_steps",
            "validation_freq",
            "class_weight",
            "max_queue_size",
            "workers",
            "use_multiprocessing",
            "initial_epoch",
        }
        super().__init__(**kwargs)

    @abc.abstractproperty
    def lookahead(self) -> int:
        """Steps ahead in y the model should target"""
        ...

    def get_metadata(self):
        """
        Add number of forecast steps to metadata

        Returns
        -------
        metadata: dict
            Metadata dictionary, including forecast steps.
        """
        metadata = super().get_metadata()
        metadata.update({"forecast_steps": self.lookahead})
        return metadata

    def _validate_and_fix_size_of_X(self, X):
        if X.ndim == 1:
            logger.info(
                f"Reshaping X from an array to an matrix of shape {(len(X), 1)}"
            )
            X = X.reshape(len(X), 1)

        if self.lookback_window >= X.shape[0]:
            raise ValueError(
                "For KerasLSTMForecast lookback_window must be < size of X"
            )
        return X

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "KerasLSTMForecast":

        """
        This fits a one step forecast LSTM architecture.

        Parameters
        ----------
        X: np.ndarray
           2D numpy array of dimension n_samples x n_features. Input data to train.
        y: np.ndarray
           2D numpy array representing the target
        kwargs: dict
            Any additional args to be passed to Keras `fit_generator` method.

        Returns
        -------
        class:
            KerasLSTMForecast

        """

        X = X.values if isinstance(X, pd.DataFrame) else X
        y = y.values if isinstance(y, pd.DataFrame) else y

        X = self._validate_and_fix_size_of_X(X)

        # We call super.fit on a single sample (notice the batch_size=1) to initiate the
        # model using the scikit-learn wrapper.
        tsg = create_keras_timeseriesgenerator(
            X=X[
                : self.lookahead + self.lookback_window
            ],  # We only need a bit of the data
            y=y[: self.lookahead + self.lookback_window],
            batch_size=1,
            lookback_window=self.lookback_window,
            lookahead=self.lookahead,
        )

        primer_x, primer_y = tsg[0]

        super().fit(X=primer_x, y=primer_y, epochs=1, verbose=0)

        tsg = create_keras_timeseriesgenerator(
            X=X,
            y=y,
            batch_size=self.batch_size,
            lookback_window=self.lookback_window,
            lookahead=self.lookahead,
        )

        gen_kwargs = {
            k: v
            for k, v in {**self.kwargs, **kwargs}.items()
            if k in self.fit_generator_params
        }

        # shuffle is set to False since we are dealing with time series data and
        # so training data will not be shuffled before each epoch.
        self.model.fit(tsg, shuffle=False, **gen_kwargs)
        return self

    def predict(self, X: np.ndarray, **kwargs) -> np.ndarray:
        """
        Parameters
        ----------
         X: np.ndarray
            Data to predict/transform. 2D numpy array of dimension `n_samples x
            n_features` where `n_samples` must be > lookback_window.

        Returns
        -------
        results: np.ndarray
                 2D numpy array of dimension `(n_samples - lookback_window) x
                 2*n_features`.  The first half of the array `(results[:,
                 :n_features])` corresponds to X offset by `lookback_window+1` (i.e.,
                 `X[lookback_window:,:]`) whereas the second half corresponds to the
                 predicted values of `X[lookback_window:,:]`.


        Example
        -------
        >>> import numpy as np
        >>> from gordo.machine.model.factories.lstm_autoencoder import lstm_model
        >>> from gordo.machine.model.models import KerasLSTMForecast
        >>> #Define train/test data
        >>> X_train = np.array([[1, 1], [2, 3], [0.5, 0.6], [0.3, 1], [0.6, 0.7]])
        >>> X_test = np.array([[2, 3], [1, 1], [0.1, 1], [0.5, 2]])
        >>> #Initiate model, fit and transform
        >>> lstm_ae = KerasLSTMForecast(kind="lstm_model",
        ...                             lookback_window=2,
        ...                             verbose=0)
        >>> model_fit = lstm_ae.fit(X_train, y=X_train.copy())
        >>> model_transform = lstm_ae.predict(X_test)
        >>> model_transform.shape
        (2, 2)
        """
        X = X.values if isinstance(X, pd.DataFrame) else X

        X = self._validate_and_fix_size_of_X(X)
        tsg = create_keras_timeseriesgenerator(
            X=X,
            y=X,
            batch_size=10000,
            lookback_window=self.lookback_window,
            lookahead=self.lookahead,
        )
        return self.model.predict(tsg)

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        """
        Returns the explained variance score between 1 step forecasted input and true
        input at next time step (note: for LSTM X is offset by `lookback_window`).

        Parameters
        ----------
        X: Union[np.ndarray, pd.DataFrame]
            Input data to the model.
        y: Union[np.ndarray, pd.DataFrame]
            Target
        sample_weight: Optional[np.ndarray]
            Sample weights

        Returns
        -------
        score: float
            Returns the explained variance score.
        """
        if not hasattr(self, "model"):
            raise NotFittedError(
                f"This {self.__class__.__name__} has not been fitted yet."
            )

        out = self.predict(X)

        # Limit X samples to match the offset causes by LSTM lookback window
        # ie, if look back window is 5, 'out' will be 5 rows less than X by now
        return explained_variance_score(y[-len(out) :], out)


class KerasLSTMForecast(KerasLSTMBaseEstimator):
    @property
    def lookahead(self) -> int:
        return 1


class KerasLSTMAutoEncoder(KerasLSTMBaseEstimator):
    @property
    def lookahead(self) -> int:
        return 0


def create_keras_timeseriesgenerator(
    X: np.ndarray,
    y: Optional[np.ndarray],
    batch_size: int,
    lookback_window: int,
    lookahead: int,
) -> tensorflow.keras.preprocessing.sequence.TimeseriesGenerator:
    """
    Provides a `keras.preprocessing.sequence.TimeseriesGenerator` for use with
    LSTM's, but with the added ability to specify the lookahead of the target in y.

    If lookahead==0 then the generated samples in X will have as their last element
    the same as the corresponding Y. If lookahead is 1 then the values in Y is shifted
    so it is one step in the future compared to the last value in the samples in X,
    and similar for larger values.


    Parameters
    ----------
    X: np.ndarray
        2d array of values, each row being one sample.
    y: Optional[np.ndarray]
        array representing the target.
    batch_size: int
        How big should the generated batches be?
    lookback_window: int
        How far back should each sample see. 1 means that it contains a single
        measurement
    lookahead: int
        How much is Y shifted relative to X

    Returns
    -------
    TimeseriesGenerator
        3d matrix with a list of batchX-batchY pairs, where batchX is a batch of
        X-values, and correspondingly for batchY. A batch consist of `batch_size` nr
        of pairs of samples (or y-values), and each sample is a list of length
        `lookback_window`.

    Examples
    -------
    >>> import numpy as np
    >>> X, y = np.random.rand(100,2), np.random.rand(100, 2)
    >>> gen = create_keras_timeseriesgenerator(X, y,
    ...                                        batch_size=10,
    ...                                        lookback_window=20,
    ...                                        lookahead=0)
    >>> len(gen) # 9 = (100-20+1)/10
    9
    >>> len(gen[0]) # batchX and batchY
    2
    >>> len(gen[0][0]) # batch_size=10
    10
    >>> len(gen[0][0][0]) # a single sample, lookback_window = 20,
    20
    >>> len(gen[0][0][0][0]) # n_features = 2
    2
    """
    new_length = len(X) + 1 - lookahead
    kwargs: Dict[str, Any] = dict(length=lookback_window, batch_size=batch_size)
    if lookahead == 1:
        kwargs.update(dict(data=X, targets=y))

    elif lookahead >= 0:

        pad_kw = dict(maxlen=new_length, dtype=X.dtype)

        if lookahead == 0:
            kwargs["data"] = pad_sequences([X], padding="post", **pad_kw)[0]
            kwargs["targets"] = pad_sequences([y], padding="pre", **pad_kw)[0]

        elif lookahead > 1:
            kwargs["data"] = pad_sequences(
                [X], padding="post", truncating="post", **pad_kw
            )[0]
            kwargs["targets"] = pad_sequences(
                [y], padding="pre", truncating="pre", **pad_kw
            )[0]
    else:
        raise ValueError(f"Value of `lookahead` can not be negative, is {lookahead}")

    return TimeseriesGenerator(**kwargs)



import logging
import inspect
from typing import Dict, Callable, Any  # pragma: no flakes
from gordo.machine.model.base import GordoBase
from tensorflow import keras

logger = logging.getLogger(__name__)


class register_model_builder:
    """
    Decorator to register a function as an available 'type' in supporting
    factory classes such as gordo_compontents.models._models.KerasAutoEncoder.

    When submitting the config file, it's important that the 'kind' is compatible
    with 'type'. 

    ie. 'type': 'KerasAutoEncoder' should support the object returned by a given
    decorated function.


    Example for KerasAutoEncoder:

    from gordo_compontents.models.register import register_model_builder

    @register_model_builder(type='KerasAutoEncoder')
    def special_keras_model_builder(n_features, ...):
        ...

    A valid yaml config would be:
    model:
        gordo.machine.models.KerasAutoEncoder:
            kind: special_keras_model_builder
    """

    """
    Mapping of type: kind: function.
    ie.
    {
        'KerasAutoEncoder' : {'special_keras_model_builder': <gordo.builder.....>},
        'ScikitRandomForest': {'special_random_forest': <gordo.build....>}
    }
    """

    factories = dict()  # type: Dict[str, Dict[str, Callable[..., keras.models.Model]]]

    def __init__(self, type: str):
        self.type = type

    def __call__(self, build_fn: Callable[..., keras.models.Model]):
        self._register(self.type, build_fn)
        return build_fn

    @classmethod
    def _register(cls, type: str, build_fn: Callable[[int, Any], GordoBase]):
        """
        Registers a given function as an available factory under
        this type.
        """
        cls._validate_func(build_fn)

        # Add function to available factories under this type
        if type not in cls.factories:
            cls.factories[type] = dict()
        cls.factories[type][build_fn.__name__] = build_fn

    @staticmethod
    def _validate_func(func):
        # Any validity checks before registering their function
        if "n_features" not in inspect.getfullargspec(func).args:
            raise ValueError(
                f"Build function: {func.__name__} does not have "
                "'n_features' as an argument; it should."
            )



# -*- coding: utf-8 -*-

import logging
import pydoc
import copy
import typing  # noqa
from typing import Union, Dict, Any, Iterable
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.base import BaseEstimator
from tensorflow.keras.models import Sequential


logger = logging.getLogger(__name__)


def import_locate(import_path: str) -> Any:
    return pydoc.locate(import_path)


def from_definition(
    pipe_definition: Union[str, Dict[str, Dict[str, Any]]]
) -> Union[FeatureUnion, Pipeline]:
    """
    Construct a Pipeline or FeatureUnion from a definition.

    Example
    -------
    >>> import yaml
    >>> from gordo import serializer
    >>> raw_config = '''
    ... sklearn.pipeline.Pipeline:
    ...         steps:
    ...             - sklearn.decomposition.PCA:
    ...                 n_components: 3
    ...             - sklearn.pipeline.FeatureUnion:
    ...                 - sklearn.decomposition.PCA:
    ...                     n_components: 3
    ...                 - sklearn.pipeline.Pipeline:
    ...                     - sklearn.preprocessing.MinMaxScaler
    ...                     - sklearn.decomposition.TruncatedSVD:
    ...                         n_components: 2
    ...             - sklearn.ensemble.RandomForestClassifier:
    ...                 max_depth: 3
    ... '''
    >>> config = yaml.safe_load(raw_config)
    >>> scikit_learn_pipeline = serializer.from_definition(config)


    Parameters
    ---------
    pipe_definition
        List of steps for the Pipeline / FeatureUnion
    constructor_class
        What to place the list of transformers into,
        either sklearn.pipeline.Pipeline/FeatureUnion

    Returns
    -------
    sklearn.pipeline.Pipeline
        pipeline
    """
    # Avoid some mutation
    definition = copy.deepcopy(pipe_definition)
    return _build_step(definition)


def _build_branch(
    definition: Iterable[Union[str, Dict[Any, Any]]],
    constructor_class=Union[Pipeline, None],
):
    """
    Builds a branch of the tree and optionally constructs the class with the given
    leafs of the branch, if constructor_class is not none. Otherwise just the
    built leafs are returned.
    """
    steps = [_build_step(step) for step in definition]
    return steps if constructor_class is None else constructor_class(steps)


def _build_scikit_branch(
    definition: Iterable[Union[str, Dict[Any, Any]]],
    constructor_class=Union[Pipeline, None],
):
    """
    Exactly like :func:`~_build_branch` except it's expected this is going to
    be a list of tuples, where the 0th element is the name of the step.
    """
    steps = [(f"step_{i}", _build_step(step)) for i, step in enumerate(definition)]
    return steps if constructor_class is None else constructor_class(steps)


def _build_step(
    step: Union[str, Dict[str, Dict[str, Any]]]
) -> Union[FeatureUnion, Pipeline, BaseEstimator]:
    """
    Build an isolated step within a transformer list, given a dict config

    Parameters
    ----------
    step: dict/str - A dict, with a single key and associated dict
                     where the associated dict are parameters for the
                     given step.

                     Example: {'sklearn.preprocessing.PCA':
                                    {'n_components': 4}
                              }
                        Gives:  PCA(n_components=4)

                    Alternatively, 'step' can be a single string, in
                    which case the step will be initiated w/ default
                    params.

                    Example: 'sklearn.preprocessing.PCA'
                        Gives: PCA()
    Returns
    -------
        Scikit-Learn Transformer or BaseEstimator
    """
    logger.debug(f"Building step: {step}")

    # Here, 'step' _should_ be a dict with a single key
    # and an associated dict containing parameters for the desired
    # sklearn step. ie. {'sklearn.preprocessing.PCA': {'n_components': 2}}
    if isinstance(step, dict):

        if len(step.keys()) != 1:
            return _load_param_classes(step)

        import_str = list(step.keys())[0]

        StepClass: Union[FeatureUnion, Pipeline, BaseEstimator] = import_locate(
            import_str
        )

        if StepClass is None:
            raise ImportError(f'Could not locate path: "{import_str}"')

        params = step.get(import_str, dict())

        if hasattr(StepClass, "from_definition"):
            return getattr(StepClass, "from_definition")(params)

        # Load any possible classes in the params if this is a dict of maybe kwargs
        if isinstance(params, dict):
            params = _load_param_classes(params)

        # update any param values which are string locations to functions
        if isinstance(params, dict):
            for param, value in params.items():
                if isinstance(value, str):
                    possible_func = import_locate(value)
                    if callable(possible_func):
                        params[param] = possible_func

        # FeatureUnion or another Pipeline transformer
        if any(StepClass == obj for obj in [FeatureUnion, Pipeline, Sequential]):

            # Need to ensure the parameters to be supplied are valid FeatureUnion
            # & Pipeline both take a list of transformers, but with different
            # kwarg, here we pull out the list to keep _build_scikit_branch generic
            if "transformer_list" in params:
                params["transformer_list"] = _build_scikit_branch(
                    params["transformer_list"], None
                )
            elif "steps" in params:
                params["steps"] = _build_scikit_branch(params["steps"], None)

            # If params is an iterable, is has to be the first argument
            # to the StepClass (FeatureUnion / Pipeline); a list of transformers
            elif any(isinstance(params, obj) for obj in (tuple, list)):
                steps = _build_scikit_branch(params, None)
                return StepClass(steps)
            elif isinstance(params, dict) and "layers" in params:
                params["layers"] = _build_branch(params["layers"], None)
            else:
                raise ValueError(
                    f"Got {StepClass} but the supplied parameters"
                    f"seem invalid: {params}"
                )
        return StepClass(**params)

    # If step is just a string, can initialize it without any params
    # ie. "sklearn.preprocessing.PCA"
    elif isinstance(step, str):
        Step = import_locate(step)  # type: Union[FeatureUnion, Pipeline, BaseEstimator]
        if hasattr(Step, "from_definition"):
            return getattr(Step, "from_definition")({})
        else:
            return Step() if Step is not None else step

    else:
        raise ValueError(
            f"Expected step to be either a string or a dict," f"found: {type(step)}"
        )


def _build_callbacks(definitions: list):
    """
    Parameters
    ----------
    definitions: List
        List of callbacks definitions

    Examples
    --------
    >>> callbacks=_build_callbacks([{'tensorflow.keras.callbacks.EarlyStopping': {'monitor': 'val_loss,', 'patience': 10}}])
    >>> type(callbacks[0])
    <class 'tensorflow.python.keras.callbacks.EarlyStopping'>

    Returns
    -------
    dict
    """
    callbacks = []
    for callback in definitions:
        callbacks.append(_build_step(callback))
    return callbacks


def _load_param_classes(params: dict):
    """
    Inspect the params' values and determine if any can be loaded as a class.
    if so, update that param's key value as the instantiation of the class.

    Additionally, if the value of the top level is a dict, and that dict's len(.keys()) == 1
    AND that key can be loaded, it's assumed to be a class whose associated values
    should be passed in as kwargs.

    Parameters
    ----------
    params: dict
        key value pairs of kwargs, which can have full class paths defined.

    Examples
    --------
    >>> params = {"key1": "value1"}
    >>> assert _load_param_classes(params) == params  # No modifications

    # Load an actual model, without any kwargs
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> params = {"base_estimator": "sklearn.ensemble.RandomForestRegressor"}
    >>> print(_load_param_classes(params))
    {'base_estimator': RandomForestRegressor()}

    # Load an actual model, with kwargs
    >>> params = {"base_estimator": {"sklearn.ensemble.RandomForestRegressor": {"n_estimators": 20}}}
    >>> print(_load_param_classes(params))
    {'base_estimator': RandomForestRegressor(n_estimators=20)}


    Returns
    -------
    dict
        Updated params which has any possible class paths loaded up as instantiated
        objects
    """
    params = copy.copy(params)
    for key, value in params.items():

        # If value is a simple string, try to load the model/class
        if isinstance(value, str):
            Model: Union[None, BaseEstimator, Pipeline] = import_locate(value)
            if Model is not None:
                if hasattr(Model, "from_definition"):
                    params[key] = getattr(Model, "from_definition")({})
                elif isinstance(Model, type) and issubclass(Model, BaseEstimator):

                    params[key] = Model()

        # For the next bit to work, the dict must have a single key (maybe) the class path,
        # and its value must be a dict of kwargs
        elif (
            isinstance(value, dict)
            and len(value.keys()) == 1
            and isinstance(value[list(value.keys())[0]], dict)
        ):
            import_path = list(value.keys())[0]
            Model = import_locate(import_path)

            sub_params = value[import_path]

            if hasattr(Model, "from_definition"):
                params[key] = getattr(Model, "from_definition")(sub_params)
            elif Model is not None and isinstance(Model, type):

                if issubclass(Model, Pipeline) or issubclass(Model, Sequential):
                    # Model is a Pipeline, so 'value' is the definition of that Pipeline
                    # Can can just re-use the entry to building a pipeline.
                    params[key] = from_definition(value)
                else:
                    # Call this func again, incase there is nested occurances of this problem in these kwargs
                    kwargs = _load_param_classes(sub_params)
                    params[key] = Model(**kwargs)  # type: ignore
        elif key == "callbacks" and isinstance(value, list):
            params[key] = _build_callbacks(value)
    return params


def load_params_from_definition(definition: dict) -> dict:
    """
    Deserialize each value from a dictionary. Could be used for preparing kwargs for methods

    Parameters
    ----------
    definition: dict
    """
    if not isinstance(definition, dict):
        raise ValueError(
            "Expected definition to be a dict," f"found: {type(definition)}"
        )
    return _load_param_classes(definition)



from .from_definition import from_definition, load_params_from_definition
from .into_definition import into_definition, load_definition_from_params
from .serializer import dump, dumps, load, loads, load_metadata
